"""
object encapsulating various optimization methods, including convex optimization

(c) Copyright Bprotocol foundation 2023. 
Licensed under MIT

NOTE: this class is not part of the API of the Carbon protocol, and you must expect breaking
changes even in minor version updates. Use at your own risk.


Convex and Marginal Price Optimization for Arbitrage and Routing
================================================================

This module implements a number of methods that allow for routing* and arbitrage amongst a set
of AMMs. Most methods allow, subject to convergence, for the optimization and routing within
an arbitrary multi-token context. The _subject to convergence_ part is important, as the in 
particular the convex optimization methods with the solvers available to us to do not seem to
be able to handle leveraged liquidity well.

This module is still subject to active research, and comments and suggestions are welcome. 
The corresponding author is Stefan Loesch <stefan@bancor.network>


*routing is not implemented yet, but it is a trivial extension of the arbitrage methods that
only needs to be connected and properly parameterized
"""
__VERSION__ = "3.6"
__DATE__ = "06/May/2023"

from dataclasses import dataclass, field, fields, asdict, astuple, InitVar
import pandas as pd
import numpy as np

try:
    import cvxpy as cp
except:
    # if cvxpy is not installed the convex optimization methods will not work; however, the
    # the marginal price based methods will still work
    cp = None
import time
import math
import numbers
import pickle
from .cpc import ConstantProductCurve as CPC, CPCInverter, CPCContainer
from sys import float_info


class _DCBase:
    """base class for all data classes, adding some useful methods"""

    def asdict(self):
        return asdict(self)

    def astuple(self):
        return astuple(self)

    def fields(self):
        return fields(self)

    # def pickle(self, filename, addts=True):
    #     """
    #     pickles the object to a file
    #     """
    #     if addts:
    #         filename = f"{filename}.{time.time()}.pickle"
    #     with open(filename, 'wb') as f:
    #         pickle.dump(self, f)

    # @classmethod
    # def unpickle(cls, filename):
    #     """
    #     unpickles the object from a file
    #     """
    #     with open(filename, 'rb') as f:
    #         object = pickle.load(f)
    #     assert isinstance(object, cls), f"unpickled object is not of type {cls}"
    #     return object


@dataclass
class ScaledVariable(_DCBase):
    """
    wraps a cvxpy variable to allow for scaling
    """

    variable: cp.Variable
    scale: any = 1.0
    token: list = None

    def __post_init__(self):
        try:
            len_var = len(self.variable.value)
        except TypeError as e:
            print("[ScaledVariable] variable.value is None", self.variable)
            return

        if not isinstance(self.scale, numbers.Number):
            self.scale = np.array(self.scale)
            if not len(self.scale) == len_var:
                raise ValueError(
                    "scale and variable must have same length or scale must be a number",
                    self.scale,
                    self.variable.value,
                )
        if not self.token is None:
            if not len(self.token) == len_var:
                raise ValueError(
                    "token and variable must have same length",
                    self.token,
                    self.variable.value,
                )

    @property
    def value(self):
        """
        converts value from USD to token units*

        Note: with scaling, the calculation is set up in a way that the values of the raw variables
        dx, dy correspond approximately to USD numbers, so their relative scale is natural and only
        determined by the problem, not by units.

        The scaling factor is the PRICE in USD PER TOKEN, therefore

            self.variable.value = USD value of the token
            self.variable.value / self.scale = number of tokens
        """
        try:
            return np.array(self.variable.value) / self.scale
        except Exception as e:
            print("[value] exception", e, self.variable.value, self.scale)
            return self.variable.value

    @property
    def v(self):
        """alias for variable"""
        return self.variable


class OptimizerBase:
    """
    base class for all optimizers

    :problem:       the problem object (eg allowing to read `problem.status`)
    :result:        the return value of problem.solve
    :time:          the time it took to solve this problem (optional)
    :optimizer:     the optimizer object that created this result
    """

    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    def pickle(self, basefilename, addts=True):
        """
        pickles the object to a file
        """
        if addts:
            filename = f"{basefilename}.{int(time.time()*100)}.optimizer.pickle"
        else:
            filename = f"{basefilename}.optimizer.pickle"
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def unpickle(cls, basefilename):
        """
        unpickles the object from a file
        """
        with open(f"{basefilename}.optimizer.pickle", "rb") as f:
            object = pickle.load(f)
        assert isinstance(object, cls), f"unpickled object is not of type {cls}"
        return object

    @dataclass
    class OptimizerResult(_DCBase):
        result: float
        time: float
        method: str = None
        optimizer: InitVar

        def __post_init__(self, optimizer):
            self._optimizer = optimizer
            # print("[OptimizerResult] post_init", optimizer)

        @property
        def optimizer(self):
            return self._optimizer

        def __float__(self):
            return float(self.result)

        @property
        def status(self):
            """problem status"""
            raise NotImplementedError("must be implemented in derived class")

        @property
        def is_error(self):
            """True if problem status is not OPTIMAL"""
            raise NotImplementedError("must be implemented in derived class")

        def detailed_error(self):
            """detailed error analysis"""
            raise NotImplementedError("must be implemented in derived class")

        @property
        def error(self):
            """problem error"""
            if not self.is_error:
                return None
            return self.detailed_error()

    @dataclass
    class SimpleResult(_DCBase):
        result: float
        method: str = None
        errormsg: str = None
        context_dct: dict = None

        def __float__(self):
            if self.is_error:
                raise ValueError("cannot convert error result to float")
            return float(self.result)

        @property
        def is_error(self):
            return not self.errormsg is None

        @property
        def context(self):
            return self.context_dct if not self.context_dct is None else {}

    DERIVEPS = 1e-6

    @classmethod
    def deriv(cls, func, x):
        """
        computes the derivative of `func` at point `x`
        """
        h = cls.DERIVEPS
        return (func(x + h) - func(x - h)) / (2 * h)

    @classmethod
    def deriv2(cls, func, x):
        """
        computes the second derivative of `func` at point `x`
        """
        h = cls.DERIVEPS
        return (func(x + h) - 2 * func(x) + func(x - h)) / (h * h)

    @classmethod
    def findmin_gd(cls, func, x0, *, learning_rate=0.1, N=100):
        """
        finds the minimum of `func` using gradient descent starting at `x0`
        """
        x = x0
        for _ in range(N):
            x -= learning_rate * cls.deriv(func, x)
        return cls.SimpleResult(result=x, method="findmin_gd")

    @classmethod
    def findmax_gd(cls, func, x0, *, learning_rate=0.1, N=100):
        """
        finds the maximum of `func` using gradient descent, starting at `x0`
        """
        x = x0
        for _ in range(N):
            x += learning_rate * cls.deriv(func, x)
        return cls.SimpleResult(result=x, method="findmax_gd")

    @classmethod
    def findminmax_nr(cls, func, x0, *, N=20):
        """
        finds the minimum or maximum of func using Newton Raphson, starting at x0
        """
        x = x0
        for _ in range(N):
            # print("[NR]", x, func(x), cls.deriv(func, x), cls.deriv2(func, x))
            try:
                x -= cls.deriv(func, x) / cls.deriv2(func, x)
            except Exception as e:
                return cls.SimpleResult(
                    result=None,
                    errormsg=f"Newton Raphson failed: {e} [x={x}, x0={x0}]",
                    method="findminmax_nr",
                )
        return cls.SimpleResult(result=x, method="findminmax_nr")

    findmin = findminmax_nr
    findmax = findminmax_nr

    GOALSEEKEPS = 1e-6

    @classmethod
    def goalseek(cls, func, a, b):
        """
        finds the value of `x` where `func(x)` x is zero, using binary search between a,b
        """
        if func(a) * func(b) > 0:
            cls.SimpleResult(
                result=None,
                errormsg=f"function must have different signs at a,b [{a}, {b}, {func(a)} {func(b)}]",
                method="findminmax_nr",
            )
            raise ValueError("function must have different signs at a,b")
        while (b - a) > cls.GOALSEEKEPS:
            c = (a + b) / 2
            if func(c) == 0:
                return c
            elif func(a) * func(c) < 0:
                b = c
            else:
                a = c
        return cls.SimpleResult(result=(a + b) / 2, method="findminmax_nr")

    @staticmethod
    def posx(vector):
        """
        returns the positive elements of the vector, zeroes elsewhere
        """
        if isinstance(vector, np.ndarray):
            return np.maximum(0, vector)
        return tuple(max(0, x) for x in vector)

    @staticmethod
    def negx(vector):
        """
        returns the negative elements of the vector, zeroes elsewhere
        """
        if isinstance(vector, np.ndarray):
            return np.minimum(0, vector)
        return tuple(min(0, x) for x in vector)

    @staticmethod
    def a(vector):
        """helper: returns vector as np.array"""
        return np.array(vector)

    @staticmethod
    def t(vector):
        """helper: returns vector as tuple"""
        return tuple(vector)

    @staticmethod
    def F(func, rg):
        """helper: returns list of [func(x) for x in rg]"""
        return [func(x) for x in rg]


FORMATTER = lambda x: "" if ((abs(x) < 1e-10) or math.isnan(x)) else f"{x:,.2f}"

F = OptimizerBase.F

TIF_OBJECTS = "objects"
TIF_DICTS = "dicts"
TIF_DFP = "dfp"
TIF_DFRAW = "dfraw"
TIF_DFAGGR = "dfaggr"
TIF_DF = "dfraw"

class CPCArbOptimizer(OptimizerBase):
    """
    main optimizer class for CPC arbitrage optimzisation
    """

    def __init__(self, curve_container):
        if not isinstance(curve_container, CPCContainer):
            curve_container = CPCContainer(curve_container)
        self._curve_container = curve_container
        
    @property
    def curve_container(self):
        """the curve container (CPCContainer)"""
        return self._curve_container

    CC = curve_container

    @property
    def tokens(self):
        return self.curve_container.tokens

    @dataclass
    class ConvexOptimizerResult(OptimizerBase.OptimizerResult):

        problem: InitVar

        def __post_init__(self, optimizer=None, problem=None, *args, **kwargs):
            super().__post_init__(*args, optimizer=optimizer, **kwargs)
            # print("[ConvexOptimizerResult] post_init")
            assert not problem is None, "problem must be set"
            self._problem = problem
            if self.method is None:
                self.method = "convex"

        @property
        def problem(self):
            return self._problem

        @property
        def status(self):
            """problem status"""
            return self.problem.status

        @property
        def is_error(self):
            """True if problem status is not OPTIMAL"""
            return self.status != cp.OPTIMAL or isinstance(self.result, str)

        @property
        def error(self):
            """problem error"""
            if not self.is_error:
                return None
            if isinstance(self.result, str):
                return f"{self.result} [{self.status}]"
            return f"{self.status}"

    @dataclass
    class NofeesOptimizerResult(ConvexOptimizerResult):
        """
        results of the nofees optimizer
        """

        token_table: dict = None
        sfc: any = field(repr=False, default=None)  # SelfFinancingConstraints
        curves: CPCContainer = field(repr=False, default=None)
        # curves_new: CPCContainer = field(repr=False,  default=None)
        # dx: cp.Variable = field(repr=False, default=None)
        # dy: cp.Variable = field(repr=False, default=None)
        dx: InitVar
        dy: InitVar

        def __post_init__(
            self, optimizer=None, problem=None, dx=None, dy=None, *args, **kwargs
        ):
            super().__post_init__(*args, optimizer=optimizer, problem=problem, **kwargs)
            # print("[NofeesOptimizerResult] post_init")
            assert not self.token_table is None, "token_table must be set"
            assert not self.sfc is None, "sfc must be set"
            assert not self.curves is None, "curves must be set"
            # assert not self.curves_new is None, "curves_new must be set"
            assert not dx is None, "dx must be set"
            assert not dy is None, "dy must be set"
            self._dx = dx
            self._dy = dy

        @property
        def dx(self):
            return self._dx

        @property
        def dy(self):
            return self._dy

        @property
        def curves_new(self):
            """returns a list of Curve objects the trade instructions implemented"""
            assert self.is_error is False, "cannot get this data from an error result"
            return self.optimizer.adjust_curves(dxvals=self.dxvalues)

        def trade_instructions(self, ti_format=None):
            """
            returns list of TradeInstruction objects

            :ti_format:     format of the TradeInstruction objects, see TradeInstruction.to_format
                            :TIF_OBJECTS:       a list of TradeInstruction objects (default)
                            :TIF_DICTS:         a list of TradeInstruction dictionaries
                            :TIF_DFRAW:         raw dataframe (holes are filled with NaN)
                            :TIF_DFP:           returns a "pretty" dataframe (holes are spaces)
                            :TIF_DFAGRR:        aggregated dataframe
                            :TIF_DF:            alias for :TIF_DFRAW:
            """
            result = (
                CPCArbOptimizer.TradeInstruction.new(
                    curve_or_cid=c, tkn1=c.tknx, amt1=dx, tkn2=c.tkny, amt2=dy
                )
                for c, dx, dy in zip(self.curves, self.dxvalues, self.dyvalues)
                if dx != 0 or dy != 0
            )
            return CPCArbOptimizer.TradeInstruction.to_format(result, ti_format)

        @property
        def dxvalues(self):
            """returns dx values"""
            return self.dx.value

        @property
        def dyvalues(self):
            """returns dy values"""
            return self.dy.value

        def dxdydf(self, *, asdict=False, pretty=True, inclk=False):
            """returns dataframe with dx, dy per curve"""
            if inclk:
                dct = [
                    {
                        "cid": c.cid,
                        "pair": c.pair,
                        "tknx": c.tknx,
                        "tkny": c.tkny,
                        "x": c.x,
                        "y": c.y,
                        "xa": c.x_act,
                        "ya": c.y_act,
                        "k": c.k,
                        "kpost": (c.x + dxv) * (c.y + dyv),
                        "kk": (c.x + dxv) * (c.y + dyv) / c.k,
                        c.tknx: dxv,
                        c.tkny: dyv,
                    }
                    for dxv, dyv, c in zip(self.dx.value, self.dy.value, self.curves)
                ]
            else:
                dct = [
                    {
                        "cid": c.cid,
                        "pair": c.pair,
                        "tknx": c.tknx,
                        "tkny": c.tkny,
                        "x": c.x,
                        "y": c.y,
                        "xa": c.x_act,
                        "ya": c.y_act,
                        "kk": (c.x + dxv) * (c.y + dyv) / c.k,
                        c.tknx: dxv,
                        c.tkny: dyv,
                    }
                    for dxv, dyv, c in zip(self.dx.value, self.dy.value, self.curves)
                ]
            if asdict:
                return dct
            df = pd.DataFrame.from_dict(dct).set_index("cid")
            df0 = df.fillna(0)
            dfa = df0[df0.columns[8:]].sum().to_frame(name="total").T
            dff = pd.concat([df, dfa], axis=0)
            if pretty:
                try:
                    dff = dff.style.format({col: FORMATTER for col in dff.columns[3:]})
                except Exception as e:
                    print("[dxdydf] exception", e, dff.columns)
            return dff

    @dataclass
    class SelfFinancingConstraints(_DCBase):
        """
        describes self financing constraints and determines optimization variable

        :data:      a dict TKN -> amount, or AMMPays, AMMReceives
                    :amount:            from the AMM perspective, total inflows (>0) or outflows (<0)
                                        for all items not present in data the value is assumed zero
                    :AMMPays:           the AMM payout should be maximized [from the trader (!) perspective]
                    :AMMReceives:       the money paid into the AMM should be minimized [ditto]
                    :OptimizationVar:   like AMMPays and AMMReceives, but if the direction of the payout is
                                        not known at the beginning [not all methods allow this]
                    :OV:                alias for OptimizationVar
        :tokens:    set of all tokens in the problem (if None, use data.keys())

        """

        AMMPays = "AMMPays"
        AMMReceives = "AMMReceives"
        OptimizationVar = "OptimizationVar"
        OV = OptimizationVar

        data: dict
        tokens: set = None

        def __post_init__(self):
            optimizationvars = tuple(
                k
                for k, v in self.data.items()
                if v in {self.AMMPays, self.AMMReceives, self.OptimizationVar}
            )
            assert (
                len(optimizationvars) == 1
            ), f"there must be EXACTLY one AMMPays, AMMReceives, OptimizationVar {self.data}"
            self._optimizationvar = optimizationvars[0]
            if self.tokens is None:
                self.tokens = set(self.data.keys())
            else:
                if isinstance(self.tokens, str):
                    self.tokens = set(t.strip() for t in self.tokens.split(","))
                else:
                    self.tokens = set(self.tokens)
                assert (
                    set(self.data.keys()) - self.tokens == set()
                ), f"constraint keys {set(self.data.keys())} > {self.tokens}"

        @property
        def optimizationvar(self):
            """optimization variable, ie the in that is set to AMMPays, AMMReceives or OptimizationVar"""
            return self._optimizationvar

        @property
        def tokens_s(self):
            """tokens as a comma-separated string"""
            return ", ".join(self.tokens_l)

        @property
        def tokens_l(self):
            """tokens as a list"""
            return sorted(list(self.tokens))

        def asdict(self, *, short=False):
            """dict representation including zero-valued tokens (unless short)"""
            if short:
                return {**self.data}
            return {k: self.get(k) for k in self.tokens}

        def items(self, *, short=False):
            return self.asdict(short=short).items()

        @classmethod
        def new(cls, tokens, **data):
            """alternative constructor: data as kwargs"""
            return cls(data=data, tokens=tokens)

        @classmethod
        def arb(cls, targettkn):
            """alternative constructor: arbitrage constraint, ie all other constraints are zero"""
            return cls(data={targettkn: cls.OptimizationVar})

        def get(self, item):
            """gets the constraint, or 0 if not present"""
            assert item in self.tokens, f"item {item} not in {self.tokens}"
            return self.data.get(item, 0)

        def is_constraint(self, item):
            """
            returns True iff item is a constraint (ie not an optimisation variable)
            """
            return not self.is_optimizationvar(item)

        def is_optimizationvar(self, item):
            """
            returns True iff item is the optimization variable
            """
            assert item in self.tokens, f"item {item} not in {self.tokens}"
            return item == self.optimizationvar

        def is_arbsfc(self):
            """
            returns True iff the constraint is an arbitrage constraint
            """
            if len(self.data) == 1:
                return True
            data1 = [v for v in self.data.values() if v != 0]
            return len(data1) == 1

        def __call__(self, item):
            """alias for get"""
            return self.get(item)

    def SFC(self, **data):
        """alias for SelfFinancingConstraints.new"""
        return self.SelfFinancingConstraints.new(self.curve_container.tokens(), **data)

    def SFCd(self, data_dct):
        """alias for SelfFinancingConstraints.new, with data as a dict"""
        return self.SelfFinancingConstraints.new(
            self.curve_container.tokens(), **data_dct
        )

    def SFCa(self, targettkn):
        """alias for SelfFinancingConstraints.arb"""
        return self.SelfFinancingConstraints.arb(targettkn)

    arb = SFCa

    AMMPays = SelfFinancingConstraints.AMMPays
    AMMReceives = SelfFinancingConstraints.AMMReceives
    OptimizationVar = SelfFinancingConstraints.OptimizationVar
    OV = SelfFinancingConstraints.OV

    SOLVER_ECOS = "ECOS"
    SOLVER_SCS = "SCS"
    SOLVER_OSQP = "OSQP"
    SOLVER_CVXOPT = "CVXOPT"
    SOLVER_CBC = "CBC"
    SOLVERS = {
        SOLVER_ECOS: cp.ECOS,
        SOLVER_SCS: cp.SCS,
        SOLVER_OSQP: cp.OSQP,
        SOLVER_CVXOPT: cp.CVXOPT,
        SOLVER_CBC: cp.CBC,
        # those solvers will usually have to be installed separately
        "ECOS_BB": cp.ECOS_BB,
        "OSQP": cp.OSQP,
        "GUROBI": cp.GUROBI,
        "MOSEK": cp.MOSEK,
        "GLPK": cp.GLPK,
        "GLPK_MI": cp.GLPK_MI,
        "CPLEX": cp.CPLEX,
        "XPRESS": cp.XPRESS,
        "SCIP": cp.SCIP,
    }

    def nofees_optimizer(self, sfc, **params):
        """
        convex optimization for determining the arbitrage opportunities

        :sfc:       a SelfFinancingConstraints object (or str passed to SFC.arb)
        :params:    additional parameters to be passed to the solver
                    :verbose:       if True, generate verbose output
                    :solver:        the solver to be used (default: "CVXOPT"; see SOLVERS)
                    :nosolve:       if True, do not solve the problem, but return the problem object
                    :nominconstr:   if True, do NOT add the minimum constraints
                    :maxconstr:     if True, DO add the (reundant) maximum constraints
                    :retcurves:     if True, also return the curves object (default: False)
                    :s_xxx:         pass the parameter `xxx` to the solver (eg s_verbose)
                    :s_verbose:     if True, generate verbose output from the solver


        note: CVXOPT is a pip install (pip install cvxopt); OSQP is not suitable for this problem,
        ECOS and SCS do work sometimes but can go dramatically wrong
        """

        # This code runs the actual optimization. It has two major parts

        # 1. the **constraints**, and
        # 2. the **objective function** to be optimized (min or max)

        # The objective function is to either maximize the number of tokens
        # received from the AMM (which is a negative number, hence formally the
        # condition is `cp.Minimize` or to minimize the number of tokens paid to
        # the AMM which is a positive number. Therefore `cp.Minimize` is the
        # correct choice in each case.

        # The constraints come in three types:

        # - **curve constraint**: the curve constraints correspond to the
        #   $x\cdot y=k$ invariant of the respective AMM; the constraint is
        #   formally `>=` but it has been shown eg by Angeris et al that the
        #   constraint will always be optimal on the boundary

        # - **range constraints**: the range constraints correspond to the
        #   tokens actually available on curve; for the full-curve AMM those
        #   constraints would formally be `dx >= -c.x` and the same for `y`, but
        #   those constraint are automatically fulfilled because of the
        #   asymptotic behaviour of the curves so could be omitted

        # - **self-financing constraints**: the self-financing constraints
        #   corresponds to the condition that all `dx` and `dy` corresponding to
        #   a specific token other than the token in the objective function must
        #   sum to the target amount provided in `inputs` (or zero if not
        #   provided)

        assert not cp is None, "cvxpy not installed [pip install cvxpy]]"
        if isinstance(sfc, str):
            sfc = self.SelfFinancingConstraints.arb(sfc)

        curves_t = self.curve_container.curves
        c0 = curves_t[0]
        tt = self.curve_container.tokentable()
        prtkn = sfc.optimizationvar

        P = lambda x: params.get(x)

        start_time = time.time()

        # set up the optimization variables
        if P("verbose"):
            print(f"Setting up dx[0..{len(curves_t)-1}] and dy[0..{len(curves_t)-1}]")
        dx = cp.Variable(len(curves_t), value=[0] * len(curves_t))
        dy = cp.Variable(len(curves_t), value=[0] * len(curves_t))

        # the geometric mean of objects in a list
        gmean = lambda lst: cp.geo_mean(cp.hstack(lst))

        ## assemble the constraints...
        constraints = []

        # curve constraints
        for i, c in enumerate(curves_t):
            constraints += [
                gmean([c.x + dx[i] / c.scalex, c.y + dy[i] / c.scaley]) >= c.kbar
            ]
            if P("verbose"):
                print(
                    f"CC {i} [{c.cid}]: {c.pair} x={c.x:.1f} {c.tknx } (s={c.scalex}), y={c.y:.1f} {c.tkny} (s={c.scaley}), k={c.k:2.1f}, p_dy/dx={c.p:2.1f}, p_dx/dy={1/c.p:2.1f}"
                )

        if P("verbose"):
            print("number of constraints: ", len(constraints))

        # range constraints (min)
        for i, c in enumerate(curves_t):

            pass

            if not P("nominconstr"):
                constraints += [
                    dx[i] / c.scalex >= c.dx_min,
                    dy[i] / c.scaley >= c.dy_min,
                ]
                if P("verbose"):
                    print(
                        f"RC {i} [{c.cid}]: dx>{c.dx_min:.4f} {c.tknx} (s={c.scalex}), dy>{c.dy_min:.4f} {c.tkny} (s={c.scaley}) [{c.pair}]"
                    )

            if P("maxconstr"):
                if not c.dx_max is None:
                    constraints += [
                        dx[i] / c.scalex <= c.dx_max,
                    ]
                if not c.dy_max is None:
                    constraints += [
                        dy[i] / c.scaley <= c.dy_max,
                    ]
                if P("verbose"):
                    print(
                        f"RC {i} [{c.cid}]: dx<{c.dx_max} {c.tknx} (s={c.scalex}), dy<{c.dy_max} {c.tkny} (s={c.scaley}) [{c.pair}]"
                    )

        if P("verbose"):
            print("number of constraints: ", len(constraints))

        # self-financing constraints
        for tkn, tknvalue in sfc.items():
            if not isinstance(tknvalue, str):
                constraints += [
                    cp.sum([dy[i] for i in tt[tkn].y])
                    + cp.sum([dx[i] for i in tt[tkn].x])
                    == tknvalue * c0.scale(tkn)
                    # note: we can access the scale from any curve as it is a class method
                ]
                if P("verbose"):
                    print(
                        f"SFC [{tkn}={tknvalue}, s={c0.scale(tkn)}]: y={[i for i in tt[tkn].y]}, x={[i for i in tt[tkn].x]}"
                    )

        if P("verbose"):
            print("number of constraints: ", len(constraints))

        # objective function  (note: AMM out is negative, AMM in is positive)
        if P("verbose"):
            print(
                f"O: y={[i for i in tt[prtkn].y]}, x={[i for i in tt[prtkn].x]}, {prtkn}"
            )

        objective = cp.Minimize(
            cp.sum([dy[i] for i in tt[prtkn].y]) + cp.sum([dx[i] for i in tt[prtkn].x])
        )

        # run the optimization
        problem = cp.Problem(objective, constraints)
        solver = self.SOLVERS.get(P("solver"), cp.CVXOPT)
        if not P("nosolve"):
            sp = {k[2:]: v for k, v in params.items() if k[:2] == "s_"}
            print("Solver params:", sp)
            if P("verbose"):
                print(f"Solving the problem with {solver}...")
            try:
                problem_result = problem.solve(solver=solver, **sp)
                # problem_result = problem.solve(solver=solver)
            except cp.SolverError as e:
                if P("verbose"):
                    print(f"Solver error: {e}")
                problem_result = str(e)
            if P("verbose"):
                print(
                    f"Problem solved in {time.time()-start_time:.2f} seconds; result: {problem_result}"
                )
        else:
            problem_result = None

        dx_ = ScaledVariable(
            dx, [c.scalex for c in curves_t], [c.tknx for c in curves_t]
        )
        dy_ = ScaledVariable(
            dy, [c.scaley for c in curves_t], [c.tkny for c in curves_t]
        )

        return self.NofeesOptimizerResult(
            problem=problem,
            sfc=sfc,
            result=problem_result,
            time=time.time() - start_time,
            dx=dx_,
            dy=dy_,
            token_table=tt,
            curves=self.curve_container,
            # curves_new=self.adjust_curves(dxvals = dx_.value),
            optimizer=self,
        )

    SO_DXDYVECFUNC = "dxdyvecfunc"
    SO_DXDYSUMFUNC = "dxdysumfunc"
    SO_DXDYVALXFUNC = "dxdyvalxfunc"
    SO_DXDYVALYFUNC = "dxdyvalyfunc"
    SO_PMAX = "pmax"
    SO_GLOBALMAX = "globalmax"
    SO_TARGETTKN = "targettkn"

    @dataclass
    class SimpleOptimizerResult(OptimizerBase.OptimizerResult):
        """
        results of the simple optimizer

        :curves:            list of curves used in the optimization, possibly wrapped in CPCInverter objects*
        :dxdyfromp_vec_f:   vector of tuples (dx, dy), as a function of p
        :dxdyfromp_sum_f:   sum of the above, also as a function of p
        :dxdyfromp_valx_f:  valx = dy/p + dx, also as a function of p
        :dxdyfromp_valy_f:  valy = dy + p*dx/p, also as a function of p
        :p_optimal:         optimal p value

        *the CPCInverter object ensures that all curves in the list correspond to the same quote
        conventions, according to the primary direction of the pair (as determined by the Pair
        object). Accordingly, tknx and tkny are always the same for all curves in the list, regardless
        of the quote direction of the pair. The CPCInverter object abstracts this away, but of course
        only for functions that are accessible through it.
        """

        NONEFUNC = lambda x: None

        curves: list = field(repr=False, default=None)
        dxdyfromp_vec_f: any = field(repr=False, default=NONEFUNC)
        dxdyfromp_sum_f: any = field(repr=False, default=NONEFUNC)
        dxdyfromp_valx_f: any = field(repr=False, default=NONEFUNC)
        dxdyfromp_valy_f: any = field(repr=False, default=NONEFUNC)
        p_optimal: float = field(repr=False, default=None)
        errormsg: str = field(repr=True, default=None)

        def __post_init__(self, *args, **kwargs):
            super().__post_init__(*args, **kwargs)
            # print("[SimpleOptimizerResult] post_init")
            assert (
                self.p_optimal is not None or self.errormsg is not None
            ), "p_optimal must be set unless errormsg is set"
            if self.method is None:
                self.method = "simple"

        @property
        def is_error(self):
            return self.errormsg is not None

        def detailed_error(self):
            return self.errormsg

        def status(self):
            return "error" if self.is_error else "converged"

        def dxdyfromp_vecs_f(self, p):
            """returns dx, dy as separate vectors instead as a vector of tuples"""
            return tuple(zip(*self.dxdyfromp_vec_f(p)))

        @property
        def tknx(self):
            return self.curves[0].tknx

        @property
        def tkny(self):
            return self.curves[0].tkny

        @property
        def tknxp(self):
            return self.curves[0].tknxp

        @property
        def tknyp(self):
            return self.curves[0].tknyp

        @property
        def pair(self):
            return self.curves[0].pair

        @property
        def pairp(self):
            return self.curves[0].pairp

        @property
        def dxdy_vecs(self):
            return self.dxdyfromp_vecs_f(self.p_optimal)

        @property
        def dxvalues(self):
            return self.dxdy_vecs[0]

        dxv = dxvalues

        @property
        def dyvalues(self):
            return self.dxdy_vecs[1]

        dyv = dyvalues

        @property
        def dxdy_vec(self):
            return self.dxdyfromp_vec_f(self.p_optimal)

        @property
        def dxdy_sum(self):
            return self.dxdyfromp_sum_f(self.p_optimal)

        @property
        def dxdy_valx(self):
            return self.dxdyfromp_valx_f(self.p_optimal)

        valx = dxdy_valx

        @property
        def dxdy_valy(self):
            return self.dxdyfromp_valy_f(self.p_optimal)

        valy = dxdy_valy

        def trade_instructions(self, ti_format=None):
            """returns list of TradeInstruction objects"""
            result = (
                CPCArbOptimizer.TradeInstruction.new(
                    curve_or_cid=c, tkn1=self.tknx, amt1=dx, tkn2=self.tkny, amt2=dy
                )
                for c, dx, dy in zip(self.curves, self.dxvalues, self.dyvalues)
                if dx != 0 or dy != 0
            )
            return CPCArbOptimizer.TradeInstruction.to_format(result, ti_format)

    def simple_optimizer(self, targettkn=None, result=None, *, params=None):
        """
        a simple optimizer that does not use cvxpy and the works only on curves on one pair

        :result:            determines what to return
                            :SO_DXDYVECFUNC:    function of p returning vector of dx,dy values
                            :SO_DXDYSUMFUNC:    function of p returning sum of dx,dy values
                            :SO_DXDYVALXFUNC:   function of p returning value of dx,dy sum in units of tknx
                            :SO_DXDYVALYFUNC:   ditto tkny
                            :SO_PMAX:           optimal p value for global max
                            :SO_GLOBALMAX:      global max of sum dx*p + dy
                            :SO_TARGETTKN:      optimizes for one token, the other is zero
        :targettkn:         token to optimize for (if result==SO_TARGETTKN); must be None if
                            result==SO_GLOBALMAX; result defaults to the corresponding value
                            depending on whether or not targettkn is None
        :params:            dict of parameters (not currently used)
        """
        start_time = time.time()
        curves_t = CPCInverter.wrap(self.curve_container)
        assert len(curves_t) > 0, "no curves found"
        c0 = curves_t[0]
        pairs = set(c.pair for c in curves_t)
        assert len(pairs) != 0, f"no pairs found, probably empty curves [{curves_t}]"
        assert (
            len(pairs) == 1
        ), f"simple_optimizer only works on curves of one pair [{pairs}]"
        assert not (
            targettkn is None and result == self.SO_TARGETTKN
        ), "targettkn must be set if result==SO_TARGETTKN"
        assert not (
            targettkn is not None and result == self.SO_GLOBALMAX
        ), f"targettkn must be None if result==SO_GLOBALMAX {targettkn}"

        dxdy = lambda r: (np.array(r[0:2]))

        dxdyfromp_vec_f = lambda p: tuple(dxdy(c.dxdyfromp_f(p)) for c in curves_t)
        if result == self.SO_DXDYVECFUNC:
            return dxdyfromp_vec_f

        dxdyfromp_sum_f = lambda p: sum(dxdy(c.dxdyfromp_f(p)) for c in curves_t)
        if result == self.SO_DXDYSUMFUNC:
            return dxdyfromp_sum_f

        dxdyfromp_valy_f = lambda p: np.dot(dxdyfromp_sum_f(p), np.array([p, 1]))
        if result == self.SO_DXDYVALYFUNC:
            return dxdyfromp_valy_f

        dxdyfromp_valx_f = lambda p: dxdyfromp_valy_f(p) / p
        if result == self.SO_DXDYVALXFUNC:
            return dxdyfromp_valx_f

        if result is None:
            if targettkn is None:
                result = self.SO_GLOBALMAX
            else:
                result = self.SO_TARGETTKN

        if not result == self.SO_TARGETTKN:
            p_avg = np.mean([c.p for c in curves_t])
            p_optimal = self.findmax(dxdyfromp_valx_f, p_avg)
            opt_result = dxdyfromp_valx_f(float(p_optimal))
            if result == self.SO_PMAX:
                return p_optimal
            elif result != self.SO_GLOBALMAX:
                raise ValueError(f"unknown result type {result}")
            method = "simple-globalmax"
        else:
            p_min = np.min([c.p for c in curves_t])
            p_max = np.max([c.p for c in curves_t])
            assert targettkn in {
                c0.tknx,
                c0.tkny,
            }, f"targettkn {targettkn} not in {c0.tknx}, {c0.tkny}"
            # we are now running a goalseek == 0 on the token that is NOT the target token
            if targettkn == c0.tknx:
                func = lambda p: dxdyfromp_sum_f(p)[1]
                p_optimal = self.goalseek(func, p_min * 0.99, p_max * 1.01)
                opt_result = dxdyfromp_sum_f(float(p_optimal))[0]
            else:
                func = lambda p: dxdyfromp_sum_f(p)[0]
                p_optimal = self.goalseek(func, p_min * 0.99, p_max * 1.01)
                opt_result = dxdyfromp_sum_f(float(p_optimal))[1]
            method = "simple-targettkn"

        if p_optimal.is_error:
            return self.SimpleOptimizerResult(
                result=None,
                time=time.time() - start_time,
                curves=curves_t,
                dxdyfromp_vec_f=dxdyfromp_vec_f,
                dxdyfromp_sum_f=dxdyfromp_sum_f,
                dxdyfromp_valx_f=dxdyfromp_valx_f,
                dxdyfromp_valy_f=dxdyfromp_valy_f,
                p_optimal=None,
                errormsg=p_optimal.errormsg,
                method=method,
                optimizer=self,
            )
        return self.SimpleOptimizerResult(
            result=opt_result,
            time=time.time() - start_time,
            curves=curves_t,
            dxdyfromp_vec_f=dxdyfromp_vec_f,
            dxdyfromp_sum_f=dxdyfromp_sum_f,
            dxdyfromp_valx_f=dxdyfromp_valx_f,
            dxdyfromp_valy_f=dxdyfromp_valy_f,
            p_optimal=float(p_optimal),
            method=method,
            optimizer=self,
        )

    def price_estimates(self, *, tknq, tknbs):
        """
        convenience function to access CPCContainer.price_estimate

        :tknq:      can only be a single token
        :tknbs:     list of tokens

        see help(CPCContainer.price_estimate) for details
        """
        return self.curve_container.price_estimates(tknqs=[tknq], tknbs=tknbs)

    JACEPS = 1e-5

    @classmethod
    def jacobian(cls, func, x, *, eps=None):
        """
        computes the Jacobian of func at point x

        :func:    a callable x=(x1..xn) -> (y1..ym), taking and returning np.arrays
        :x:       a vector x=(x1..xn) as np.array
        """
        if eps is None:
            eps = cls.JACEPS
        n = len(x)
        y = func(x)
        jac = np.zeros((n, n))
        for j in range(n):  # through columns to allow for vector addition
            Dxj = abs(x[j]) * eps if x[j] != 0 else eps
            x_plus = [(xi if k != j else xi + Dxj) for k, xi in enumerate(x)]
            jac[:, j] = (func(x_plus) - y) / Dxj
        return jac

    J = jacobian

    MO_DEBUG = "debug"
    MO_PSTART = "pstart"
    MO_P = MO_PSTART
    MO_DTKNFROMPF = "dtknfrompf"
    MO_MINIMAL = "minimal"
    MO_FULL = "full"

    MOEPS = 1e-6
    MOMAXITER = 50
    
    class OptimizationError(Exception): pass
    class ConvergenceError(OptimizationError): pass
    class ParameterError(OptimizationError): pass

    def margp_optimizer(self, sfc=None, result=None, *, params=None):
        """
        optimal transactions across all curves in the optimizer, extracting targettkn*

        :sfc:               the self financing constraint to use**
        :result:            the result type
                            :MO_DEBUG:         a number of items useful for debugging
                            :MO_PSTART:        price estimates (as dataframe)
                            :MO_PE:            alias for MO_ESTPRICE
                            :MO_DTKNFROMPF:    the function calculating dtokens from p
                            :MO_MINIMAL:       minimal result (omitting some big fields)
                            :MO_FULL:          full result
                            :None:             alias for MO_FULL
        :params:            dict of parameters
                            :eps:              precision parameter for accepting the result (default: 1e-6)
                            :maxiter:          maximum number of iterations (default: 100)
                            :verbose:          if True, print some high level output
                            :progress:         if True, print some basic progress output
                            :debug:            if True, print some debug output
                            :debug2:           more debug output
                            :raiseonerror:     if True, raise an OptimizationError exception on error
                            :pstart:           starting price for optimization, either as dict {tkn:p, ...},
                                                or as df as price estimate as returned by MO_PSTART;
                                                excess tokens can be provided but all required tokens must be present

        :returns:           MargpOptimizerResult on the default path, others depending on the
                            chosen result

        *this optimizer uses the marginal price method, ie it solves the equation

            dx_i (p) = 0 for all i != targettkn, and the whole price vector

        **at the moment only the trivial self-financing constraint is allowed, ie the one that
        only specifies the target token, and where all other constraints are zero; if sfc is
        a string then this is interpreted as the target token
        """
        # data conversion: string to SFC object; note that anything but pure arb not currently supported
        if isinstance(sfc, str):
            sfc = self.arb(targettkn=sfc)
        assert sfc.is_arbsfc(), "only pure arbitrage SFC are supported at the moment"
        targettkn = sfc.optimizationvar
        
        # lambdas
        P      = lambda item: params.get(item, None) if params is not None else None
        get    = lambda p, ix: p[ix] if ix is not None else 1       # safe get from tuple
        dxdy_f = lambda r: (np.array(r[0:2]))                       # extract dx, dy from result
        tn     = lambda t: t.split("-")[0]                          # token name, eg WETH-xxxx -> WETH
        
        # initialisations
        eps = P("eps") or self.MOEPS
        maxiter = P("maxiter") or self.MOMAXITER
        start_time = time.time()
        curves_t = self.curve_container
        alltokens_s = self.curve_container.tokens()
        tokens_t = tuple(t for t in alltokens_s if t != targettkn) # all _other_ tokens...
        tokens_ix = {t: i for i, t in enumerate(tokens_t)}         # ...with index lookup
        pairs = self.curve_container.pairs(standardize=False)
        curves_by_pair = { 
            pair: tuple(c for c in curves_t if c.pair == pair) for pair in pairs }
        pairs_t = tuple(tuple(p.split("/")) for p in pairs)
        
        try:
        
            # assertions
            if len (curves_t) == 0:
                raise self.ParameterError("no curves found")
            if len (curves_t) == 1:
                raise self.ParameterError(f"can't run arbitrage on single curve {curves_t}")
            if not targettkn in alltokens_s:
                raise self.ParameterError(f"targettkn {targettkn} not in {alltokens_s}")
                
            # calculating the start price for the iteration process
            if not P("pstart") is None:
                pstart = P("pstart")
                if P("verbose") or P("debug"):
                    print(f"[margp_optimizer] using pstartd [{len(P('pstart'))} tokens]")
                if isinstance(P("pstart"), pd.DataFrame):
                    try:
                        pstart = pstart.to_dict()[targettkn]
                    except Exception as e:
                        raise Exception(
                            f"error while converting dataframe pstart to dict: {e}",
                            pstart,
                            targettkn,
                        )
                    assert isinstance(
                        pstart, dict
                    ), f"pstart must be a dict or a data frame [{pstart}]"
                price_estimates_t = tuple(pstart[t] for t in tokens_t)
            else:
                if P("verbose") or P("debug"):
                    print("[margp_optimizer] calculating price estimates")
                try:
                    price_estimates_t = self.price_estimates(tknq=targettkn, tknbs=tokens_t)
                except Exception as e:
                    if P("verbose") or P("debug"):
                        print(
                            "[margp_optimizer] error while calculating price estimates:", e
                        )
                    price_estimates_t = None
            if P("debug"):
                print("[margp_optimizer] pstart:", price_estimates_t)
            if result == self.MO_PSTART:
                df = pd.DataFrame(price_estimates_t, index=tokens_t, columns=[targettkn])
                df.index.name = "tknb"
                return df
            
            ## INNER FUNCTION: CALCULATE THE TARGET FUNCTION
            def dtknfromp_f(p, *, islog10=True, asdct=False):
                """
                calculates the aggregate change in token amounts for a given price vector

                :p:         price vector, where prices use the reference token as quote token
                            this vector is an np.array, and the token order is the same as in tokens_t
                :islog10:   if True, p is interpreted as log10(p)
                :asdct:     if True, the result is returned as dict AND tuple, otherwise as np.array
                :returns:   if asdct is False, a tuple of the same length as tokens_t detailing the
                            change in token amounts for each token except for the target token (ie the
                            quantity with target zero; if asdct is True, that same information is
                            returned as dict, including the target token.
                """
                p = np.array(p, dtype=np.float64)
                if islog10:
                    p = np.exp(p * np.log(10))
                assert len(p) == len(
                    tokens_t
                ), f"p and tokens_t have different lengths [{p}, {tokens_t}]"
                if P("debug"):
                    print(f"\n[dtknfromp_f] =====================>>>")
                    print(f"prices={p}")
                    print(f"tokens={tokens_t}")

                sum_by_tkn = {t: 0 for t in alltokens_s}
                for pair, (tknb, tknq) in zip(pairs, pairs_t):
                    price = get(p, tokens_ix.get(tknb)) / get(p, tokens_ix.get(tknq))
                    curves = curves_by_pair[pair]
                    c0 = curves[0]
                    dxdy = tuple(dxdy_f(c.dxdyfromp_f(price)) for c in curves)
                    if P("debug2"):
                        print(f"\n{c0.pairp} --->>")
                        print(f"  price={price:,.4f}, 1/price={1/price:,.4f}")
                        for r, c in zip(dxdy, curves):
                            s = f"  cid={c.cid:15}"
                            s += f" dx={float(r[0]):15,.3f} {c.tknxp:>5}"
                            s += f" dy={float(r[1]):15,.3f} {c.tknyp:>5}"
                            s += f" p={c.p:,.2f} 1/p={1/c.p:,.2f}"
                            print(s)
                        print(f"<<--- {c0.pairp}")

                    sumdx, sumdy = sum(dxdy)
                    sum_by_tkn[tknq] += sumdy
                    sum_by_tkn[tknb] += sumdx

                    if P("debug"):
                        print(
                            f"pair={c0.pairp}, {sumdy:,.4f} {tn(tknq)}, {sumdx:,.4f} {tn(tknb)}, price={price:,.4f} {tn(tknq)} per {tn(tknb)} [{len(curves)} funcs]"
                        )

                result = tuple(sum_by_tkn[t] for t in tokens_t)
                if P("debug"):
                    print(f"sum_by_tkn={sum_by_tkn}")
                    print(f"result={result}")
                    print(f"<<<===================== [dtknfromp_f]")

                if asdct:
                    return sum_by_tkn, np.array(result)

                return np.array(result)
            ## END INNER FUNCTION

            # return the inner function if requested
            if result == self.MO_DTKNFROMPF:
                return dtknfromp_f

            # return debug info if requested
            if result == self.MO_DEBUG:
                return dict(
                    # price_estimates_all = price_estimates_all,
                    # price_estimates_d = price_estimates_d,
                    price_estimates_t=price_estimates_t,
                    tokens_t=tokens_t,
                    tokens_ix=tokens_ix,
                    pairs=pairs,
                    sfc=sfc,
                    targettkn=targettkn,
                    pairs_t=pairs_t,
                    dtknfromp_f=dtknfromp_f,
                    optimizer=self,
                )


            # setting up the optimization variables (note: we optimize in log space)
            if price_estimates_t is None:
                raise Exception(f"price estimates not found; try setting pstart")
            p = np.array(price_estimates_t, dtype=float)
            plog10 = np.log10(p)
            if P("verbose"):
                # dtkn_d, dtkn = dtknfromp_f(plog10, islog10=True, asdct=True)
                print("[margp_optimizer] pe  ", p)
                print("[margp_optimizer] p   ", ", ".join(f"{x:,.2f}" for x in p))
                print("[margp_optimizer] 1/p ", ", ".join(f"{1/x:,.2f}" for x in p))
                # print("[margp_optimizer] dtkn", dtkn)
                # if P("tknd"):
                #     print("[margp_optimizer] dtkn_d", dtkn_d)

            ## MAIN OPTIMIZATION LOOP
            for i in range(maxiter):

                if P("progress"):
                    print(
                        f"Iteration [{i:2.0f}]: time elapsed: {time.time()-start_time:.2f}s"
                    )

                # calculate the change in token amounts (also as dict if requested)
                if P("tknd"):
                    dtkn_d, dtkn = dtknfromp_f(plog10, islog10=True, asdct=True)
                else:
                    dtkn = dtknfromp_f(plog10, islog10=True, asdct=False)

                # calculate the Jacobian
                if P("debug"):
                    print("\n[margp_optimizer] ============= JACOBIAN =============>>>")
                J = self.J(dtknfromp_f, plog10)  
                    # ATTENTION: dtknfromp_f takes log10(p) as input
                if P("debug"):
                    print("==== J ====>")
                    print(J)
                    print("<=== J =====")
                    print("<<<============= JACOBIAN ============= [margp_optimizer]\n")
                
                # Update p, dtkn using the Newton-Raphson formula
                try:
                    dplog10 = np.linalg.solve(J, -dtkn)
                except np.linalg.LinAlgError:
                    if P("verbose") or P("debug"):
                        print("[margp_optimizer] singular Jacobian, using lstsq instead")
                    dplog10 = np.linalg.lstsq(J, -dtkn, rcond=None)[0]
                    # https://numpy.org/doc/stable/reference/generated/numpy.linalg.solve.html
                    # https://numpy.org/doc/stable/reference/generated/numpy.linalg.lstsq.html
                
                # update log prices, prices and determine the criterium...
                p0log10 = [*plog10]
                plog10 += dplog10
                p = np.exp(plog10 * np.log(10))
                criterium = np.linalg.norm(dplog10)
                
                # ...print out some info if requested...
                if P("verbose"):
                    print(f"\n[margp_optimizer] ========== cycle {i} =======>>>")
                    print("log p0", p0log10)
                    print("log dp", dplog10)
                    print("log p ", plog10)
                    print("p     ", tuple(p))
                    print("p     ", ", ".join(f"{x:,.2f}" for x in p))
                    print("1/p   ", ", ".join(f"{1/x:,.2f}" for x in p))
                    print("tokens_t", tokens_t)
                    # print("dtkn", dtkn)
                    print("dtkn", ", ".join(f"{x:,.3f}" for x in dtkn))
                    print(
                        f"[criterium={criterium:.2e}, eps={eps:.1e}, c/e={criterium/eps:,.0e}]"
                    )
                    if P("tknd"):
                        print("dtkn_d", dtkn_d)
                    if P("J"):
                        print("J", J)
                    print(f"<<<========== cycle {i} ======= [margp_optimizer]")

                # ...and finally check the criterium (percentage changes this step) for convergence
                if criterium < eps:
                    if i != 0:
                        # we don't break in the first iteration because we need this first iteration
                        # to establish a common baseline price, therefore d logp ~ 0 is not good
                        # in the first step                      
                        break
                    # else:
                    #     # we break in the first loop, so we restore the initial price estimates
                    #     # (if we do log10 / 10**p then we get results that are slightly off zero)
                    #     p = np.array(price_estimates_t, dtype=float)
            ## END MAIN OPTIMIZATION LOOP
            
            if i >= maxiter - 1:
                raise self.ConvergenceError(f"maximum number of iterations reached [{i}]")
            
            NOMR = lambda f: f if not result == self.MO_MINIMAL else None
                # this function screens out certain results when MO_MINIMAL [minimal output] is chosen
            dtokens_d, dtokens_t = dtknfromp_f(p, asdct=True, islog10=False)
            return self.MargpOptimizerResult(
                optimizer=NOMR(self),
                result=dtokens_d[targettkn],
                time=time.time() - start_time,
                targettkn=targettkn,
                curves=NOMR(curves_t),
                p_optimal=NOMR({tkn: p_ for tkn, p_ in zip(tokens_t, p)}),
                p_optimal_t=tuple(p),
                dtokens=NOMR(dtokens_d),
                dtokens_t=tuple(dtokens_t),
                tokens_t=tokens_t,
                n_iterations=i,
            )
        
        except self.OptimizationError as e:
            if P("debug") or P("verbose"):
                print(f"[margp_optimizer] exception occured {e}")
                
            if P("raiseonerror"):
                raise
            
            NOMR = lambda f: f
            return self.MargpOptimizerResult(
                optimizer=NOMR(self),
                result=None,
                time=time.time() - start_time,
                targettkn=targettkn,
                curves=NOMR(curves_t),
                p_optimal=None,
                p_optimal_t=None,
                dtokens=None,
                dtokens_t=None,
                tokens_t=tokens_t,
                n_iterations=i,
                errormsg=e,
            )

    @dataclass
    class TradeInstruction(_DCBase):
        """
        encodes a trade

        seen from the AMM; in numbers must be positive, out numbers negative
        """

        cid: any
        tknin: str
        amtin: float
        tknout: str
        amtout: float
        error: str = field(repr=True, default=None)
        curve: InitVar = None
        raiseonerror: InitVar = False

        POSNEGEPS = 1e-8

        def __post_init__(self, curve=None, raiseonerror=False):
            self.curve = curve
            if curve is not None:
                if self.cid != curve.cid:
                    err = f"curve/cid mismatch [{self.cid} vs {curve.cid}]"
                    self.error = err
                    if raiseonerror:
                        raise ValueError(err)
            if self.tknin == self.tknout:
                err = f"tknin and tknout must be different [{self.tknin} {self.tknout}]"
                self.error = err
                if raiseonerror:
                    raise ValueError(err)
            self.cid = str(self.cid)
            self.tknin = str(self.tknin)
            self.tknout = str(self.tknout)
            self.amtin = float(self.amtin)
            self.amtout = float(self.amtout)
            if not self.amtin * self.amtout < 0:
                if (
                    abs(self.amtin) < self.POSNEGEPS
                    and abs(self.amtout) < self.POSNEGEPS
                ):
                    self.amtin = 0
                    self.amtout = 0
                else:
                    err = f"amtin and amtout must be of different sign [{self.amtin} {self.tknin}, {self.amtout} {self.tknout}]"
                    self.error = err
                    if raiseonerror:
                        raise ValueError(err)

            if not self.amtin >= 0:
                err = f"amtin must be positive [{self.amtin}]"  # seen from AMM
                self.error = err
                if raiseonerror:
                    raise ValueError(err)
            
            if not self.amtout <= 0:
                err = f"amtout must be negative [{self.amtout}]"  # seen from AMM
                self.error = err
                if raiseonerror:
                    raise ValueError(err)

        TIEPS = 1e-10

        @classmethod
        def new(cls, curve_or_cid, tkn1, amt1, tkn2, amt2, *, eps=None, raiseonerror=False):
            """automatically determines which is in and which is out"""
            try:
                cid = curve_or_cid.cid
                curve = curve_or_cid
            except:
                cid = curve_or_cid
                curve = None
            if eps is None:
                eps = cls.TIEPS
            if amt1 > 0:
                newobj = cls(
                    cid=cid,
                    tknin=tkn1,
                    amtin=amt1,
                    tknout=tkn2,
                    amtout=amt2,
                    curve=curve,
                    raiseonerror=raiseonerror,
                )
            else:
                newobj = cls(
                    cid=cid,
                    tknin=tkn2,
                    amtin=amt2,
                    tknout=tkn1,
                    amtout=amt1,
                    curve=curve,
                    raiseonerror=raiseonerror,
                )

            return newobj

        @property
        def is_empty(self):
            """returns True if this is an empty trade instruction (too close to zero)"""
            return self.amtin == 0 or self.amtout == 0

        @classmethod
        def to_dicts(cls, trade_instructions):
            """converts iterable ot TradeInstruction objects to a list of dicts"""
            return [ti.asdict() for ti in trade_instructions]

        @classmethod
        def to_df(cls, trade_instructions, ti_format=None):
            """converts iterable ot TradeInstruction objects to a pandas dataframe"""
            if ti_format is None:
                ti_format = cls.TIF_DF
            dicts = (
                {
                    "cid": ti.cid,
                    "pair": ti.curve.pair if not ti.curve is None else "",
                    "pairp": ti.curve.pairp if not ti.curve is None else "",
                    "tknin": ti.tknin,
                    "tknout": ti.tknout,
                    ti.tknin: ti.amtin,
                    ti.tknout: ti.amtout,
                }
                for ti in trade_instructions
            )
            df = pd.DataFrame.from_dict(list(dicts)).set_index("cid")
            if ti_format == cls.TIF_DFRAW:
                return df
            if ti_format == cls.TIF_DFAGGR:
                df1r = df[df.columns[4:]]
                df1 = df1r.fillna(0)
                dfa = df1.sum().to_frame(name="TOTAL NET").T
                dfp = df1[df1 > 0].sum().to_frame(name="AMMIn").T
                dfn = df1[df1 < 0].sum().to_frame(name="AMMOut").T
                return pd.concat([df1r, dfp, dfn, dfa], axis=0)
                return df1, dfa
            if ti_format == cls.TIF_DFP:
                return df.fillna("")
            raise ValueError(f"unknown format {ti_format}")

        TIF_OBJECTS = TIF_OBJECTS
        TIF_DICTS = TIF_DICTS
        TIF_DFP = TIF_DFP
        TIF_DFRAW = TIF_DFRAW
        TIF_DFAGGR = TIF_DFAGGR
        TIF_DF = TIF_DF

        @classmethod
        def to_format(cls, trade_instructions, ti_format=None):
            """converts iterable ot TradeInstruction objects to the given format (TIF_XXX)"""
            if ti_format is None:
                ti_format = cls.TIF_OBJECTS
            if ti_format == cls.TIF_OBJECTS:
                return tuple(trade_instructions)
            elif ti_format == cls.TIF_DICTS:
                return cls.to_dicts(trade_instructions)
            elif ti_format[:2] == "df":
                trade_instructions = tuple(trade_instructions)
                if len(trade_instructions) == 0:
                    return pd.DataFrame()
                return cls.to_df(trade_instructions, ti_format=ti_format)
            else:
                raise ValueError(f"unknown format {ti_format}")

        @property
        def price_outperin(self):
            return -self.amtout / self.amtin

        p = price_outperin

        @property
        def price_inperout(self):
            return -self.amtin / self.amtout

        pr = price_inperout

        @property
        def prices(self):
            return (self.price_outperin, self.price_inperout)

        pp = prices

    TIF_OBJECTS = TIF_OBJECTS
    TIF_DICTS = TIF_DICTS
    TIF_DFP = TIF_DFP
    TIF_DFRAW = TIF_DFRAW
    TIF_DFAGGR = TIF_DFAGGR
    TIF_DF = TIF_DF
    
    @dataclass
    class MargpOptimizerResult(OptimizerBase.OptimizerResult):
        """
        results of the simple optimizer

        :p_optimal:         optimal p values

        """
        TIF_OBJECTS = TIF_OBJECTS
        TIF_DICTS = TIF_DICTS
        TIF_DFP = TIF_DFP
        TIF_DFRAW = TIF_DFRAW
        TIF_DFAGGR = TIF_DFAGGR
        TIF_DF = TIF_DF

        curves: list = field(repr=False, default=None)
        targettkn: str = field(repr=True, default=None)
        p_optimal: dict = field(repr=False, default=None)
        p_optimal_t: tuple = field(repr=True, default=None)
        n_iterations: int = field(repr=False, default=None)
        dtokens: dict = field(repr=False, default=None)
        dtokens_t: tuple = field(repr=True, default=None)
        tokens_t: tuple = field(repr=True, default=None)
        errormsg: str = field(repr=True, default=None)

        def __post_init__(self, *args, **kwargs):
            super().__post_init__(*args, **kwargs)
            # #print("[MargpOptimizerResult] post_init")
            assert (
                self.p_optimal_t is not None or self.errormsg is not None
            ), "p_optimal_t must be set unless errormsg is set"
            if self.method is None:
                self.method = "margp"
            self.raiseonerror = False

        @property
        def is_error(self):
            return self.errormsg is not None

        def detailed_error(self):
            return self.errormsg

        def status(self):
            return "error" if self.is_error else "converged"

        def price(self, tknb, tknq):
            """returns the optimal price of tknb/tknq based on p_optimal [in tknq per tknb]"""
            assert (
                self.p_optimal is not None
            ), "p_optimal must be set [do not use minimal results]"
            return self.p_optimal.get(tknb, 1) / self.p_optimal.get(tknq, 1)

        def dxdyvalues(self, asdict=False):
            """
            returns a vector of (dx, dy) values for each curve
            """
            assert (
                self.curves is not None
            ), "curves must be set [do not use minimal results]"
            assert self.is_error is False, "cannot get this data from an error result"
            result = (
                (c.cid, c.dxdyfromp_f(self.price(c.tknb, c.tknq))[0:2])
                for c in self.curves
            )
            if asdict:
                return {cid: dxdy for cid, dxdy in result}
            return tuple(dxdy for cid, dxdy in result)

        @property
        def dxvalues(self):
            return tuple(dx for dx, dy in self.dxdyvalues())

        @property
        def dyvalues(self):
            return tuple(dy for dx, dy in self.dxdyvalues())

        @property
        def curves_new(self):
            """returns a list of Curve objects the trade instructions implemented"""
            assert (
                self.optimizer is not None
            ), "optimizer must be set [do not use minimal results]"
            assert self.is_error is False, "cannot get this data from an error result"
            return self.optimizer.adjust_curves(dxvals=self.dxvalues)

        def trade_instructions(self, ti_format=None):
            """
            returns list of TradeInstruction objects
            
            :ti_format:     TIF_OBJECTS, TIF_DICTS, TIF_DFP, TIF_DFRAW, TIF_DFAGGR, TIF_DF
            """
            try:
                assert self.curves is not None, "curves must be set [do not use minimal results]"
                assert self.is_error is False, "cannot get this data from an error result"
                result = (
                    CPCArbOptimizer.TradeInstruction.new(
                        curve_or_cid=c, tkn1=c.tknx, amt1=dx, tkn2=c.tkny, amt2=dy
                    )
                    for c, dx, dy in zip(self.curves, self.dxvalues, self.dyvalues)
                    if dx != 0 or dy != 0
                )
                return CPCArbOptimizer.TradeInstruction.to_format(result, ti_format)
            except AssertionError:
                if self.raiseonerror:
                    raise
                return None

    def adjust_curves(self, dxvals, *, verbose=False, raiseonerror=False):
        """
        returns a new curve container with the curves shifted by the given dx values
        """
        # print("[adjust_curves]", dxvals)
        if dxvals is None:
            if raiseonerror:
                raise ValueError("dxvals is None")
            else:
                print("[adjust_curves] dxvals is None")
                return None
        curves = self.curve_container
        try:
            newcurves = [
                c.execute(dx=dx, verbose=verbose, ignorebounds=True)
                for c, dx in zip(curves, dxvals)
            ]
            return CPCContainer(newcurves)
        except Exception as e:
            if raiseonerror:
                raise e
            else:
                print(f"Error in adjust_curves: {e}")
                # raise e
                return None

    def plot(self, *args, **kwargs):
        """
        convenience for self.curve_container.plot()

        see help(CPCContainer.plot) for details
        """
        return self.curve_container.plot(*args, **kwargs)

    def format(self, *args, **kwargs):
        """
        convenience for self.curve_container.format()

        see help(CPCContainer.format) for details
        """
        return self.curve_container.format(*args, **kwargs)




"""
estimating the scale of the token price in USD

NOTE: this class is not part of the Bancor Simulator API, and breaking changes may occur at any time

(c) Copyright Bprotocol foundation 2023. 
Licensed under MIT
"""
__VERSION__ = "1.0"
__DATE__ = "07/Apr/2022"

from dataclasses import dataclass, field, asdict, InitVar


class TokenScaleBase:
    """
    the "scale" of a token, ie the number of tokens per USD, typically rounded to the next power of 10
    """

    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    DEFAULT_SCALE = 1e-2

    def scale(self, token):
        """
        returns the scale of the token*

        :tkn: the token whose scale is to be returned

        *the "scale" of a token the number of tokens per USD, typically rounded
        to the next power of 10; every token MUST have a scale; if the _scale
        function (that implements this method) returns None, DEFAULT_SCALE is returned
        """
        result = self._scale(token)
        if result is None:
            result = self.DEFAULT_SCALE
        return result

    def _scale(self, token):
        """
        implements the scale method in derived classes
        """
        raise NotImplementedError("{self.__class__.__name__} did not implement _scale")

    def __call__(self, token):
        """alias for scale"""
        return self.scale(token)


class TokenScale1(TokenScaleBase):
    """trivial implementation of TokenScaleBase returning unit scale for all tokens"""

    DEFAULT_SCALE = 1e00

    def _scale(self, token):
        """implementation of _scale for TokenScale1 class: always returns unit scale"""
        return self.DEFAULT_SCALE


@dataclass
class TokenScale(TokenScaleBase):
    """
    implements the `TokenScaleBase` interface using a dictionary
    """

    scale_dct: dict = field(default_factory=dict)

    @classmethod
    def from_tokenscales(cls, **scale):
        """alternative constructor with the scale in kwargs"""
        return cls(scale_dct=scale)

    def add_scale(self, token, scale):
        """
        adds (or replaces) a scale for a token
        """
        self.scale_dct[token] = scale

    def _scale(self, token):
        """implementation of _scale for TokenScale class (reading from dict)"""
        return self.scale_dct.get(token, None)


TokenScaleData = TokenScale.from_tokenscales(
    USDC=1e00,
    USDT=1e00,
    LINK=1e01,
    AAVE=1e02,
    ETH=1e03,
    WETH=1e03,
    WBTC=1e04,
    BTC=1e04,
    BNT=1e00,
    SUSHI=1e00,
    UNI=1e01,
)

TokenScale1Data = TokenScale1()



"""
cloaking access to python files

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
__VERSION__ = "1.0-BETA1"
__DATE__ = "26/Apr 2023"

from dataclasses import dataclass

@dataclass
class ShieldedAttribute():
    """returned instead of the attribute value for shielded attributes"""
    attr: str = None
    exists: bool = None
    
class Cloaker():
    """
    Cloaking access to python files -- Base class
    
    :cloaked_item:    the item to cloak
    
    USAGE
    
    This class contains two methods, plus the associated implementation methods
    that can be overridden in subclasses, that determine its behaviour:
    
    - is_visible:   if the attribute exists and visible it is passed through
    - is_shielded:  this is called if either the attribute does not exist, or it is
                    not visible (second parameter); if it returns True then the
                    result is provided as None, otherwise an AttributeError is raised
    """
    def __init__(self, cloaked_item):
        self._cloaked_item = cloaked_item
        
    def _is_visible(self, attr_name):
        """
        whether the attribute is visible (wrapper; calls ___is_visible)
        
        :attr_name:    name of the attribute to test
        :returns:      True if visible, False otherwise
        """
        return self._is_visible_(attr_name)
    
    def _is_visible_(self, attr_name):
        """
        implementation function of _is_visible [CloakerBase]
        
        visible iff not starting with "_"
        """
        #print("[Cloaker:_is_visible_]", attr_name)
        return not attr_name.startswith("_")
    
    def _is_shielded(self, attr_name, attr_exists):
        """
        whether the (invisible) attribute is None (wrapper; calls ___is_shielded)
        
        :attr_name:     name of the attribute to test
        :attr_exists:   whether the attribute exists on the cloaked item
                        (if it exists then visible = False)
        :returns:       True if None, False otherwise
        """
        return self._is_shielded_(attr_name, attr_exists)
    
    def _is_shielded_(self, attr_name, attr_exists):
        """
        implementation function of _is_shielded [CloakerBase]
        
        None iff it exists
        """
        #print("[__is_shielded]", attr_name)
        return attr_exists
    
    _ShieldedAttribute = ShieldedAttribute
    _SHIELDF = ShieldedAttribute
    def __getattr__(self, attr_name):
        if hasattr(self._cloaked_item, attr_name):
            #print("[CloakerBase] has attribute", attr_name)
            if self._is_visible(attr_name):
                return getattr(self._cloaked_item, attr_name)
            if self._is_shielded(attr_name, True):
                return self._SHIELDF(attr_name, True)
        else:
            #print("[CloakerBase] does NOT have attribute", attr_name)
            if self._is_shielded(attr_name, False):
                return self._SHIELDF(attr_name, False)
        raise AttributeError(f"Cloaked[{self._cloaked_item.__class__.__name__}] has no attribute '{attr_name}'")

class CloakerL(Cloaker):
    """Cloaker with (visible) set of attributes"""
    def __init__(self, cloaked_item, visible):
        super().__init__(cloaked_item)
        if isinstance(visible, str):
            visible = (x.strip() for x in visible.split(","))
        self._visible = set(visible)
        
    def _is_visible_(self, attr_name):
        """
        visible iff in visible set
        """
        #print("[CloakerL:_is_visible_]", attr_name)
        return attr_name in self._visible
    
    def _is_shielded_(self, attr_name, attr_exists):
        """always shielded"""
        return True


# coding=utf-8
"""
This module contains the tests for the manager class methods which are multicall related
"""

from unittest.mock import Mock, patch, call

import brownie
import pytest
from unittest.mock import MagicMock
from brownie import multicall as brownie_multicall

from fastlane_bot import Bot, Config
from fastlane_bot.events.exchanges import (
    UniswapV2,
    UniswapV3,
    SushiswapV2,
    CarbonV1,
    BancorV3,
)
from fastlane_bot.events.managers.base import Manager
from fastlane_bot.tools.cpc import ConstantProductCurve as CPC

print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Bot))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV3))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SushiswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonV1))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(BancorV3))
from fastlane_bot.testing import *

plt.style.use("seaborn-dark")
plt.rcParams["figure.figsize"] = [12, 6]
from fastlane_bot import __VERSION__

require("3.0", __VERSION__)


@pytest.fixture
def pool_data():
    nan = np.NaN
    pool_data = [
        {
            "cid": "0xb3b0dbb95f1f70e1f053360d9bccef3fbe7c5e2b615e833a9faae18c299a0fc9",
            "last_updated": nan,
            "last_updated_block": 17634372,
            "descr": "bancor_v3 BNT-FF1C/MATIC-eBB0 0.000",
            "pair_name": "BNT-FF1C/MATIC-eBB0",
            "exchange_name": "bancor_v3",
            "fee": "0.000",
            "fee_float": 0.0,
            "address": "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
            "anchor": nan,
            "tkn0_address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "tkn1_address": "0x7D1AfA7B718fb893dB30A3aBc0Cfc608AaCfeBB0",
            "tkn0_key": "BNT-FF1C",
            "tkn1_key": "MATIC-eBB0",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "exchange_id": 2,
            "tkn0_symbol": "BNT",
            "tkn1_symbol": "MATIC",
            "timestamp": nan,
            "tkn0_balance": 371729474548077247680443,
            "tkn1_balance": 216554584335493056216168,
            "liquidity": nan,
            "sqrt_price_q96": nan,
            "tick": nan,
            "tick_spacing": 0,
        },
        {
            "cid": "0x38d373a29b8a7e621ee373ee76138184a67092259bd24ab1434ec90b98235efd",
            "last_updated": nan,
            "last_updated_block": 17634372,
            "descr": "bancor_v3 BNT-FF1C/ENS-9D72 0.000",
            "pair_name": "BNT-FF1C/ENS-9D72",
            "exchange_name": "bancor_v3",
            "fee": "0.000",
            "fee_float": 0.0,
            "address": "0xBC19712FEB3a26080eBf6f2F7849b417FdD792CA",
            "anchor": nan,
            "tkn0_address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "tkn1_address": "0xC18360217D8F7Ab5e7c516566761Ea12Ce7F9D72",
            "tkn0_key": "BNT-FF1C",
            "tkn1_key": "ENS-9D72",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "exchange_id": 2,
            "tkn0_symbol": "BNT",
            "tkn1_symbol": "ENS",
            "timestamp": nan,
            "tkn0_balance": 104058085529730176588006,
            "tkn1_balance": 4547023863756451207684,
            "liquidity": nan,
            "sqrt_price_q96": nan,
            "tick": nan,
            "tick_spacing": 0,
        },
        {
            "cid": "0x56f1f774ece226fac7c9940c98ead630bfc18a39fa2f2bdcdc56e6234d4477b1",
            "last_updated": nan,
            "last_updated_block": 17634372,
            "descr": "bancor_v3 BNT-FF1C/ALPHA-0975 0.000",
            "pair_name": "BNT-FF1C/ALPHA-0975",
            "exchange_name": "bancor_v3",
            "fee": "0.000",
            "fee_float": 0.0,
            "address": "0x8798249c2E607446EfB7Ad49eC89dD1865Ff4272",
            "anchor": nan,
            "tkn0_address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "tkn1_address": "0xa1faa113cbE53436Df28FF0aEe54275c13B40975",
            "tkn0_key": "BNT-FF1C",
            "tkn1_key": "ALPHA-0975",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "exchange_id": 2,
            "tkn0_symbol": "BNT",
            "tkn1_symbol": "ALPHA",
            "timestamp": nan,
            "tkn0_balance": 0,
            "tkn1_balance": 0,
            "liquidity": nan,
            "sqrt_price_q96": nan,
            "tick": nan,
            "tick_spacing": 0,
        },
        {
            "cid": "0x1b65e937b57a618d40da4236ae854d33a843042a9abc84ba72d808ad67435a42",
            "last_updated": nan,
            "last_updated_block": 17634372,
            "descr": "bancor_v3 BNT-FF1C/HEGIC-8430 0.000",
            "pair_name": "BNT-FF1C/HEGIC-8430",
            "exchange_name": "bancor_v3",
            "fee": "0.000",
            "fee_float": 0.0,
            "address": "0x5218E472cFCFE0b64A064F055B43b4cdC9EfD3A6",
            "anchor": nan,
            "tkn0_address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "tkn1_address": "0x584bC13c7D411c00c01A62e8019472dE68768430",
            "tkn0_key": "BNT-FF1C",
            "tkn1_key": "HEGIC-8430",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "exchange_id": 2,
            "tkn0_symbol": "BNT",
            "tkn1_symbol": "HEGIC",
            "timestamp": nan,
            "tkn0_balance": 0,
            "tkn1_balance": 0,
            "liquidity": nan,
            "sqrt_price_q96": nan,
            "tick": nan,
            "tick_spacing": 0,
        },
        {
            "cid": "0x561b7f22cadc1359057c07c5a1f11ae4d087a753aa87629ed92b38175e60c3ae",
            "last_updated": nan,
            "last_updated_block": 17634372,
            "descr": "bancor_v3 BNT-FF1C/ZCN-3B78 0.000",
            "pair_name": "BNT-FF1C/ZCN-3B78",
            "exchange_name": "bancor_v3",
            "fee": "0.000",
            "fee_float": 0.0,
            "address": "0x1559FA1b8F28238FD5D76D9f434ad86FD20D1559",
            "anchor": nan,
            "tkn0_address": "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "tkn1_address": "0xb9EF770B6A5e12E45983C5D80545258aA38F3B78",
            "tkn0_key": "BNT-FF1C",
            "tkn1_key": "ZCN-3B78",
            "tkn0_decimals": 18,
            "tkn1_decimals": 10,
            "exchange_id": 2,
            "tkn0_symbol": "BNT",
            "tkn1_symbol": "ZCN",
            "timestamp": nan,
            "tkn0_balance": 109709381661658805787397,
            "tkn1_balance": 3264962522647673,
            "liquidity": nan,
            "sqrt_price_q96": nan,
            "tick": nan,
            "tick_spacing": 0,
        },
    ]
    return pool_data


@pytest.fixture
def carbon_events():
    carbon_events = [
        {
            "last_updated_block": 17634377,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn1_address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "tkn0_symbol": "ETH",
            "tkn1_symbol": "DAI",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "cid": 340282366920938463463374607431768211457,
            "tkn0_key": "ETH-EEeE",
            "tkn1_key": "DAI-1d0F",
            "pair_name": "ETH-EEeE/DAI-1d0F",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 ETH-EEeE/DAI-1d0F 0.002",
            "y_0": 1000000000000000,
            "y_1": 0,
            "z_0": 1000000000000000,
            "z_1": 0,
            "A_0": 0,
            "A_1": 0,
            "B_0": 6382340776412,
            "B_1": 1875443170982464,
        },
        {
            "last_updated_block": 17634377,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "tkn1_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn0_symbol": "DAI",
            "tkn1_symbol": "ETH",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "cid": 340282366920938463463374607431768211593,
            "tkn0_key": "DAI-1d0F",
            "tkn1_key": "ETH-EEeE",
            "pair_name": "DAI-1d0F/ETH-EEeE",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 DAI-1d0F/ETH-EEeE 0.002",
            "y_0": 0,
            "y_1": 0,
            "z_0": 0,
            "z_1": 0,
            "A_0": 88739322630080,
            "A_1": 30784910546,
            "B_0": 1876725096051745,
            "B_1": 6418617024516,
        },
        {
            "last_updated_block": 17634377,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn1_address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "tkn0_symbol": "ETH",
            "tkn1_symbol": "DAI",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "cid": 340282366920938463463374607431768211614,
            "tkn0_key": "ETH-EEeE",
            "tkn1_key": "DAI-1d0F",
            "pair_name": "ETH-EEeE/DAI-1d0F",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 ETH-EEeE/DAI-1d0F 0.002",
            "y_0": 157076304796171508,
            "y_1": 191076681422897394849,
            "z_0": 257505273765924104,
            "z_1": 462002783910000000000,
            "A_0": 197764438468,
            "A_1": 235894416821184,
            "B_0": 6293971818901,
            "B_1": 1873305839476414,
        },
        {
            "last_updated_block": 17634377,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn1_address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "tkn0_symbol": "ETH",
            "tkn1_symbol": "DAI",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "cid": 340282366920938463463374607431768211607,
            "tkn0_key": "ETH-EEeE",
            "tkn1_key": "DAI-1d0F",
            "pair_name": "ETH-EEeE/DAI-1d0F",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 ETH-EEeE/DAI-1d0F 0.002",
            "y_0": 0,
            "y_1": 0,
            "z_0": 0,
            "z_1": 0,
            "A_0": 69065909368,
            "A_1": 106270057837888,
            "B_0": 6457478834827,
            "B_1": 1874403645842739,
        },
        {
            "last_updated_block": 17634377,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn1_address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "tkn0_symbol": "ETH",
            "tkn1_symbol": "DAI",
            "tkn0_decimals": 18,
            "tkn1_decimals": 18,
            "cid": 340282366920938463463374607431768211673,
            "tkn0_key": "ETH-EEeE",
            "tkn1_key": "DAI-1d0F",
            "pair_name": "ETH-EEeE/DAI-1d0F",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 ETH-EEeE/DAI-1d0F 0.002",
            "y_0": 1,
            "y_1": 940344,
            "z_0": 9403439,
            "z_1": 940344,
            "A_0": 0,
            "A_1": 0,
            "B_0": 785475461108442,
            "B_1": 2814749767,
        },
    ]
    return carbon_events


@pytest.fixture
def manager(pool_data):
    cfg = Config.new(config=Config.CONFIG_MAINNET)
    w3 = cfg.w3
    manager = Manager(
        cfg=cfg,
        pool_data=pool_data,
        alchemy_max_block_fetch=20,
        web3=w3,
        SUPPORTED_EXCHANGES=["bancor_v3", "carbon_v1"],
    )
    return manager


def get_rows_to_update(pool_data):
    # return the index of the pool_data
    rows_to_update = [idx for idx, row in enumerate(pool_data)]
    return rows_to_update


# This is a simple counter variable that we'll increment every time we call the function.
multicall_counter = 0


def my_multicall(*args, **kwargs):
    # This is a wrapper function that increments the counter each time it's called,
    # then calls the original function.
    global multicall_counter
    multicall_counter += 1
    return brownie_multicall(*args, **kwargs)


@patch.object(brownie, "multicall", autospec=True)
@patch("fastlane_bot.events.manager.Base.update")
def test_update_pools_directly_from_contracts_bancor_v3(
    self, update_mock, manager, pool_data
):

    # Replace the original function with our wrapper function.
    brownie.multicall = my_multicall

    # Now when you call your function, the counter will be incremented each time
    # the brownie.multicall function is called.
    manager.update_pools_directly_from_contracts(
        n_jobs=2, rows_to_update=[0, 1, 2, 3], not_bancor_v3=False, current_block=1
    )

    # And you can check the value of the counter.
    assert multicall_counter == 1

    # Check the 'update' function calls
    expected_calls = [
        call(pool_info=pool_data[i], limiter=False, block_number=1)
        for i in [0, 1, 2, 3]
    ]
    manager.update.assert_has_calls(expected_calls, any_order=True)


@patch.object(brownie, "multicall", autospec=True)  # We mock the multicall function
def test_get_strats_by_pair(self, manager):

    manager.cfg.MULTICALL_CONTRACT_ADDRESS = (
        "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696"
    )
    all_pairs = [(1, 2), (3, 4)]
    carbon_controller_mock = Mock()
    carbon_controller_mock.strategiesByPair.side_effect = [[(5, 6)], [(7, 8)]]
    result = manager.get_strats_by_pair(all_pairs, carbon_controller_mock)
    carbon_controller_mock.strategiesByPair.assert_has_calls(
        [call(*pair) for pair in all_pairs]
    )
    brownie.multicall.assert_called_once_with(
        address="0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696"
    )



"""
Fastlane bot configuration object -- selector constants
"""

NETWORK_ETHEREUM = "ethereum"
NETWORK_MAINNET = NETWORK_ETHEREUM
NETWORK_TENDERLY = "tenderly"

DATABASE_SQLITE = "sqlite"
DATABASE_POSTGRES = "postgres"
DATABASE_MEMORY = "memory"
DATABASE_SDK = "sdk"
DATABASE_UNITTEST = "unittest"

LOGGER_DEFAULT = "default"

LOGLEVEL_DEBUG = "debug"
LOGLEVEL_INFO = "info"
LOGLEVEL_WARNING = "warning"
LOGLEVEL_ERROR = "error"

PROVIDER_DEFAULT = "default"  # the default provider for network
PROVIDER_INFURA = "infura"
PROVIDER_ALCHEMY = "alchemy"
PROVIDER_TENDERLY = "tenderly"
PROVIDER_UNITTEST = "unittest"



"""
Fastlane bot configuration object -- main object
"""

__VERSION__ = "1.0"
__DATE__ = "03/May 2023"

import os
from dataclasses import dataclass, field, InitVar, asdict
# from .base import ConfigBase
from . import network as network_, db as db_, logger as logger_, provider as provider_
from .cloaker import CloakerL
from . import selectors as S
from dotenv import load_dotenv

from .connect import EthereumNetwork

load_dotenv()
TENDERLY_FORK_ID = os.environ.get("TENDERLY_FORK_ID")
if TENDERLY_FORK_ID is None:
    TENDERLY_FORK_ID = ''
WEB3_ALCHEMY_PROJECT_ID = os.environ.get("WEB3_ALCHEMY_PROJECT_ID")
PROVIDER_URL = f'https://rpc.tenderly.co/fork/{TENDERLY_FORK_ID}' if TENDERLY_FORK_ID != '' else f"https://eth-mainnet.alchemyapi.io/v2/{WEB3_ALCHEMY_PROJECT_ID}"
NETWORK_ID = 'mainnet' if TENDERLY_FORK_ID == '' else 'tenderly'
NETWORK_NAME = "Ethereum Mainnet" if TENDERLY_FORK_ID == '' else 'Tenderly (Alchemy)'


@dataclass
class Config():
    """
    Fastlane bot configuration object
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    network: network_.ConfigNetwork = field(default=None)
    db: db_.ConfigDB = field(default=None)
    logger: logger_.ConfigLogger = field(default=None)
    provider: provider_.ConfigProvider = field(default=None)

    CONFIG_UNITTEST = "unittest"
    CONFIG_TENDERLY = "tenderly"
    CONFIG_MAINNET = "mainnet"

    LOGLEVEL_DEBUG = S.LOGLEVEL_DEBUG
    LOGLEVEL_INFO = S.LOGLEVEL_INFO
    LOGLEVEL_WARNING = S.LOGLEVEL_WARNING
    LOGLEVEL_ERROR = S.LOGLEVEL_ERROR

    LL_DEBUG = S.LOGLEVEL_DEBUG
    LL_INFO = S.LOGLEVEL_INFO
    LL_WARN = S.LOGLEVEL_WARNING
    LL_ERR = S.LOGLEVEL_ERROR

    SUPPORTED_EXCHANGES = ['carbon_v1', 'bancor_v2', 'bancor_v3', 'uniswap_v2', 'uniswap_v3', 'sushiswap_v2']
    connection = EthereumNetwork(
        network_id=NETWORK_ID,
        network_name=NETWORK_NAME,
        provider_url=PROVIDER_URL,
        provider_name="alchemy",
    )
    connection.connect_network()
    w3 = connection.web3

    @classmethod
    def new(cls, *, config=None, loglevel=None, **kwargs):
        """
        Alternative constructor: create and return new Config object
        
        :config:    CONFIG_MAINNET(default), CONFIG_TENDERLY, CONFIG_UNITTEST
        :loglevel:  LOGLEVEL_DEBUG, LOGLEVEL_INFO (default), LOGLEVEL_WARNING, LOGLEVEL_ERROR
        """
        if config is None:
            config = cls.CONFIG_MAINNET

        if loglevel is None:
            loglevel = cls.LOGLEVEL_INFO
        C_log = logger_.ConfigLogger.new(loglevel=loglevel)

        if config == cls.CONFIG_MAINNET:
            C_nw = network_.ConfigNetwork.new(network=S.NETWORK_MAINNET)
            return cls(network=C_nw, logger=C_log, **kwargs)
        elif config == cls.CONFIG_TENDERLY:
            C_db = db_.ConfigDB.new(db=S.DATABASE_POSTGRES, POSTGRES_DB="tenderly")
            C_nw = network_.ConfigNetwork.new(network=S.NETWORK_TENDERLY)
            return cls(db=C_db, logger=C_log, network=C_nw, **kwargs)
        elif config == cls.CONFIG_UNITTEST:
            C_db = db_.ConfigDB.new(db=S.DATABASE_UNITTEST, POSTGRES_DB="unittest")
            C_nw = network_.ConfigNetwork.new(network=S.NETWORK_MAINNET)
            C_pr = provider_.ConfigProvider.new(network=C_nw, provider=S.PROVIDER_DEFAULT)
            return cls(db=C_db, logger=C_log, network=C_nw, provider=C_pr, **kwargs)
        raise ValueError(f"Invalid config: {config}")

    def is_config_item(self, item):
        """returns True if item is a (possible) configuration item [uppercase, numbers, underscore; len>2]"""
        # print("[is_config_item]", item)
        if item in {"w3", "connection"}:
            return True
        if len(item) < 3:
            return False
        if not item[0].isupper():
            return False
        for c in item[1:]:
            if not (item.isupper() or item.isnumeric() or item == "_"):
                return False
        return True

    def get_attribute_from_config(self, name: str):
        """
        gets the attribute from the constituent config objects, raises if not found
        """
        for obj in [self.network, self.db, self.provider, self.logger]:
            if hasattr(obj, name):
                return getattr(obj, name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __getattr__(self, name: str):
        """
        If of type attribute, return it.
        """
        if self.is_config_item(name):
            return self.get_attribute_from_config(name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __post_init__(self):
        """
        Post-initialization initialization.
        """
        if self.network is None:
            self.network = network_.ConfigNetwork.new(network_.ConfigNetwork.NETWORK_ETHEREUM)
        assert issubclass(type(self.network), network_.ConfigNetwork)

        if self.db is None:
            self.db = db_.ConfigDB.new(db_.ConfigDB.DATABASE_POSTGRES)
        assert issubclass(type(self.db), db_.ConfigDB)

        # if self.fastlane is None:
        #     self.fastlane = fastlane_.ConfigFastlane.new()
        # assert issubclass(type(self.fastlane), fastlane_.ConfigFastlane)

        if self.logger is None:
            self.logger = logger_.ConfigLogger.new()
        assert issubclass(type(self.logger), logger_.ConfigLogger)

        if self.provider is None:
            self.provider = provider_.ConfigProvider.new(self.network)
        assert issubclass(type(self.provider), provider_.ConfigProvider)

        assert self.network is self.provider.network, f"Network mismatch: {self.network} != {self.provider.network}"

    VISIBLE_FIELDS = "network, db, logger, provider, w3, ZERO_ADDRESS"

    def cloaked(self, incl=None, excl=None):
        """
        returns a cloaked version of the object
        
        :incl:  fields to _include_ in the cloaked version (plus those in VISIBLE_FIELDS)
        :excl:  fields to _exclude_ from the cloaked version
        """
        visible = self.VISIBLE_FIELDS
        if isinstance(visible, str):
            visible = (x.strip() for x in visible.split(","))
        visible = set(visible)

        if isinstance(incl, str):
            incl = (x.strip() for x in incl.split(","))
        elif incl is None:
            incl = []
        visible |= set(incl)

        if isinstance(excl, str):
            excl = (x.strip() for x in excl.split(","))
        elif excl is None:
            excl = []
        visible -= set(excl)

        return CloakerL(self, visible)



# coding=utf-8
"""
Contains the base class for the managers modules.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import time
from dataclasses import dataclass, field
from typing import List, Dict, Any, Type, Optional, Tuple

import brownie
from web3 import Web3
from web3.contract import Contract

from fastlane_bot import Config
from fastlane_bot.events.exchanges import exchange_factory
from fastlane_bot.events.exchanges.base import Exchange
from fastlane_bot.events.pools import pool_factory


@dataclass
class BaseManager:
    """
    The Base class is responsible for coordinating the data fetching process.

    Parameters
    ----------
    web3 : Web3
        The Web3 instance.
    cfg : Config
        The Config instance.
    pool_data : List[Dict[str, Any]]
        The pool data.
    alchemy_max_block_fetch : int
        The maximum number of blocks to fetch from Alchemy.
    event_contracts : Dict[str, Contract or Type[Contract]]
        The event contracts.
    pool_contracts : Dict[str, Contract or Type[Contract]]
        The pool contracts.
    erc20_contracts : Dict[str, Contract or Type[Contract]]
        The ERC20 contracts.
    exchanges : Dict[str, Exchange]
        The exchanges.
    uniswap_v2_event_mappings : Dict[str, str]
        The UniswapV2 event mappings.
    unmapped_uni2_events : List[str]
        The unmapped UniswapV2 events.
    tokens : List[Dict[str, str]]
        The tokens.
    TOKENS_MAPPING : Dict[str, Any]
        The tokens mapping.
    SUPPORTED_EXCHANGES : Dict[str, Any]
        The supported exchanges.
    """

    web3: Web3
    cfg: Config
    pool_data: List[Dict[str, Any]]
    alchemy_max_block_fetch: int
    event_contracts: Dict[str, Contract or Type[Contract]] = field(default_factory=dict)
    pool_contracts: Dict[str, Contract or Type[Contract]] = field(default_factory=dict)
    erc20_contracts: Dict[str, Contract or Type[Contract]] = field(default_factory=dict)
    exchanges: Dict[str, Exchange] = field(default_factory=dict)
    uniswap_v2_event_mappings: Dict[str, str] = field(default_factory=dict)
    unmapped_uni2_events: List[str] = field(default_factory=list)
    tokens: List[Dict[str, str]] = field(default_factory=dict)

    TOKENS_MAPPING: Dict[str, Any] = field(
        default_factory=lambda: {
            "ETH_ADDRESS": ("ETH", 18),
            "WETH_ADDRESS": ("WETH", 18),
            "WBTC_ADDRESS": ("WBTC", 8),
            "BNT_ADDRESS": ("BNT", 18),
            "USDC_ADDRESS": ("USDC", 6),
        }
    )

    SUPPORTED_EXCHANGES: List[str] = None

    def __post_init__(self):
        for exchange_name in self.SUPPORTED_EXCHANGES:
            self.exchanges[exchange_name] = exchange_factory.get_exchange(exchange_name)
        self.init_exchange_contracts()

    @staticmethod
    def exchange_name_from_event(event: Dict[str, Any]) -> str:
        """
        Get the exchange name from the event.

        Parameters
        ----------
        event : Dict[str, Any]
            The event.

        Returns
        -------
        str
            The exchange name.
        """
        return next(
            (
                exchange_name
                for exchange_name, pool_class in pool_factory._creators.items()
                if pool_class.event_matches_format(event)
            ),
            None,
        )

    def check_forked_exchange_names(
        self, exchange_name_default: str = None, address: str = None, event: Any = None
    ) -> str:
        """
        Check the forked exchange names. If the exchange name is forked (Sushiswap from UniswapV2, etc) return the
        real exchange name.

        Parameters
        ----------
        exchange_name_default : str, optional
            The default exchange name.
        address : str, optional
            The address.
        event : Any, optional
            The event.

        Returns
        -------
        str
            The real exchange name.

        """
        if exchange_name_default is None:
            exchange_name_default = self.exchange_name_from_event(event)

        return exchange_name_default

    def get_tkn_info(self, address: str) -> Tuple[Optional[str], Optional[int]]:
        """
        Get the token info.

        Parameters
        ----------
        address : str
            The address.

        Returns
        -------
        Tuple[Optional[str], Optional[int]]
            The token info.

        """
        tkns = self.get_tkn_symbol_and_decimals(
            self.web3,
            self.erc20_contracts,
            self.cfg,
            self.web3.toChecksumAddress(address),
        )
        return tkns or (None, None)

    def multicall(self, address):
        return brownie.multicall(address)

    def get_rows_to_update(self, update_from_contract_block: int) -> List[int]:
        """
        Get the rows to update.

        Parameters
        ----------
        update_from_contract_block : int
            The latest block number to update from.

        Returns
        -------
        List[int]
            The rows to update.

        """

        if "carbon_v1" in self.SUPPORTED_EXCHANGES:
            start_time = time.time()
            self.cfg.logger.info("Updating carbon pools w/ multicall...")

            # Create a CarbonController contract object
            carbon_controller = brownie.Contract.from_abi(
                address=self.cfg.CARBON_CONTROLLER_ADDRESS,
                abi=self.exchanges["carbon_v1"].get_abi(),
                name="CarbonController",
            )

            # Store the contract object in pool_contracts
            self.pool_contracts["carbon_v1"][
                self.cfg.CARBON_CONTROLLER_ADDRESS
            ] = carbon_controller

            # Create a list of pairs from the CarbonController contract object
            pairs = [(second, first) for first, second in carbon_controller.pairs()]

            # Create a reversed list of pairs
            pairs_reverse = [(second, first) for first, second in pairs]

            # Combine both pair lists and add extra parameters
            all_pairs = [(pair[0], pair[1], 0, 5000) for pair in pairs]

            strategies_by_pair = self.get_strats_by_pair(all_pairs, carbon_controller)

            # expand strategies_by_pair
            strategies_by_pair = [
                s for strat in strategies_by_pair if strat for s in strat if s
            ]

            # Log the time taken for the above operations
            self.cfg.logger.info(
                f"Fetched {len(strategies_by_pair)} carbon strategies in {time.time() - start_time} seconds"
            )

            start_time = time.time()
            current_block = self.web3.eth.blockNumber

            # Create pool info for each strategy
            for strategy in strategies_by_pair:
                if len(strategy) > 0:
                    self.add_pool_info_from_event(
                        strategy=strategy, block_number=current_block
                    )

            # Log the time taken for the above operations
            self.cfg.logger.info(
                f"Updated {len(strategies_by_pair)} carbon strategies info in {time.time() - start_time} seconds"
            )

        return [
            i
            for i, pool_info in enumerate(self.pool_data)
            if pool_info["last_updated_block"]
            < update_from_contract_block - self.alchemy_max_block_fetch
        ]

    def get_strats_by_pair(self, all_pairs, carbon_controller):
        with self.multicall(address=self.cfg.MULTICALL_CONTRACT_ADDRESS):
            # Fetch strategies for each pair from the CarbonController contract object
            strategies_by_pair = [
                carbon_controller.strategiesByPair(*pair) for pair in all_pairs
            ]
        return strategies_by_pair

    def get_tkn_symbol_and_decimals(
        self, web3: Web3, erc20_contracts: Dict[str, Contract], cfg: Config, addr: str
    ) -> Tuple[str, int]:
        """
        Get the token symbol and decimals.

        Parameters
        ----------
        web3 : Web3
            The web3 instance.
        erc20_contracts : Dict[str, Contract]
            The erc20 contracts.
        cfg : Config
            The config.
        addr : str
            The address.

        Returns
        -------
        Tuple[str, int]
            The token symbol and decimals.

        """
        token_info = self.get_token_info_from_config(cfg, addr)
        if token_info:
            return token_info

        record = next((add for add in self.tokens if add["address"] == addr), None)
        if record:
            return record["symbol"], int(record["decimals"])

        return self.get_token_info_from_contract(web3, erc20_contracts, addr)

    def get_token_info_from_config(
        self, cfg: Config, addr: str
    ) -> Optional[Tuple[str, int]]:
        """
        Get the token info from config.

        Parameters
        ----------
        cfg : Config
            The config.
        addr : str
            The address.

        Returns
        -------
        Optional[Tuple[str, int]]
            The token info.

        """
        return next(
            (
                info
                for address_attr, info in self.TOKENS_MAPPING.items()
                if addr in [getattr(cfg, address_attr)]
            ),
            None,
        )

    def validate_pool_info(
        self,
        addr: Optional[str] = None,
        event: Optional[Dict[str, Any]] = None,
        pool_info: Optional[Dict[str, Any]] = None,
        key: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Validate the pool info.

        Parameters
        ----------
        addr : Optional[str], optional
            The address, by default None
        event : Optional[Dict[str, Any]], optional
            The event, by default None
        pool_info : Optional[Dict[str, Any]], optional
            The pool info, by default None
        key : Optional[str], optional
            The key, by default None
        block_number : Optional[int], optional
            The block number, by default None

        Returns
        -------
        Optional[Dict[str, Any]]
            The pool info.

        """
        if key != "cid" and (pool_info is None or not pool_info):
            pool_info = self.add_pool_info_from_contract(
                address=addr, event=event, block_number=event["blockNumber"]
            )

        if addr == self.cfg.CARBON_CONTROLLER_ADDRESS:
            cid = event["args"]["id"] if event is not None else pool_info["cid"]
            for pool in self.pool_data:
                if pool["cid"] == cid:
                    pool_info = pool
                    break

        if isinstance(pool_info, float):
            return

        return pool_info

    def get_key_and_value(
        self, event: Dict[str, Any], addr: str, ex_name: str
    ) -> Tuple[str, Any]:
        """
        Get the key and value.

        Parameters
        ----------
        event : Dict[str, Any]
            The event.
        addr : str
            The address.
        ex_name : str
            The exchange name.

        Returns
        -------
        Tuple[str, Any]
            The key and value.

        """
        if ex_name == "carbon_v1":
            return "cid", event["args"]["id"]
        if ex_name in {"uniswap_v2", "sushiswap_v2", "uniswap_v3"}:
            return "address", addr
        if ex_name == "bancor_v3":
            value = (
                event["args"]["tkn_address"]
                if event["args"]["tkn_address"] != self.cfg.BNT_ADDRESS
                else event["args"]["pool"]
            )
            return "tkn1_address", value

    def handle_strategy_deleted(self, event: Dict[str, Any]) -> None:
        """
        Handle the strategy deleted event.

        Parameters
        ----------
        event : Dict[str, Any]
            The event.
        """
        cid = event["args"]["id"]
        self.pool_data = [p for p in self.pool_data if p["cid"] != cid]
        self.exchanges["carbon_v1"].delete_strategy(event["args"]["id"])

    def deduplicate_pool_data(self) -> None:
        """
        Deduplicate the pool data.
        """
        self.pool_data = sorted(
            self.pool_data, key=lambda x: x["last_updated_block"], reverse=True
        )
        seen = set()
        no_duplicates = [
            d for d in self.pool_data if d["cid"] not in seen and not seen.add(d["cid"])
        ]
        self.pool_data = no_duplicates

    @staticmethod
    def pool_key_value_from_event(key: str, event: Dict[str, Any]) -> Any:
        """
        Get the pool key value from the event.

        Parameters
        ----------
        key : str
            The key.
        event : Dict[str, Any]
            The event.

        Returns
        -------
        Any
            The pool key value.
        """
        if key == "cid":
            return event["args"]["id"]
        elif key == "address":
            return event["address"]
        elif key == "tkn0_address":
            return event["args"]["token0"]
        elif key == "tkn1_address":
            return event["args"]["token1"]

    print_events = []




__VERSION__ = "0.0"
__DATE__="01/May/2023"

class TxReceiptHandlerBase:
    __VERSION__=__VERSION__
    __DATE__=__DATE__

class TxReceiptHandler(TxReceiptHandlerBase):
    __VERSION__=__VERSION__
    __DATE__=__DATE__


# coding=utf-8
"""
This module contains the tests for the exchanges classes
"""
from fastlane_bot import Bot, Config
from fastlane_bot.bot import CarbonBot
from fastlane_bot.tools.cpc import ConstantProductCurve as CPC
from fastlane_bot.events.exchanges import UniswapV2, UniswapV3, SushiswapV2, CarbonV1, BancorV3
from fastlane_bot.events.interface import QueryInterface
from fastlane_bot.helpers.poolandtokens import PoolAndTokens
import pytest

print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Bot))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV3))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SushiswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonV1))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(BancorV3))
from fastlane_bot.testing import *

plt.style.use('seaborn-dark')
plt.rcParams['figure.figsize'] = [12,6]
from fastlane_bot import __VERSION__
require("3.0", __VERSION__)

# Setup

@pytest.fixture(scope="module")
def setup_bot():
    C = Config.new(config=Config.CONFIG_MAINNET)
    assert C.NETWORK == C.NETWORK_MAINNET
    assert C.PROVIDER == C.PROVIDER_ALCHEMY
    setup_bot = CarbonBot(ConfigObj=C)
    pools = [PoolAndTokens(ConfigObj=C, id=87, cid=12590447576074723148144860474975423823949, last_updated=None, last_updated_block=17632852, descr='carbon_v1 BAM-4AaB/ETH-EEeE 0.002', pair_name='BAM-4AaB/ETH-EEeE', exchange_name='carbon_v1', fee='0.002', fee_float=0.002, tkn0_balance=None, tkn1_balance=None, z_0=0, y_0=0, A_0=0, B_0=5232231541229414, z_1=0, y_1=0, A_1=46278427, B_1=1688849860, sqrt_price_q96=None, tick=None, tick_spacing=None, liquidity=None, address='0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1', anchor=None, tkn0='BAM', tkn1='ETH', tkn0_address='0x9DB0FB0Aebe6A925B7838D16e3993A3976A64AaB', tkn0_decimals=18, tkn1_address='0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE', tkn1_decimals=18, tkn0_key='BAM-4AaB', tkn1_key='ETH-EEeE'), PoolAndTokens(ConfigObj=C, id=88, cid=12930729942995661611608235082407192035408, last_updated=None, last_updated_block=17632852, descr='carbon_v1 stETH-fE84/WETH-6Cc2 0.002', pair_name='stETH-fE84/WETH-6Cc2', exchange_name='carbon_v1', fee='0.002', fee_float=0.002, tkn0_balance=None, tkn1_balance=None, z_0=0, y_0=0, A_0=11537531850305, B_0=267164224749167, z_1=1000000000000000000, y_1=1000000000000000000, A_1=4298139552137, B_1=274347871120865, sqrt_price_q96=None, tick=None, tick_spacing=None, liquidity=None, address='0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1', anchor=None, tkn0='stETH', tkn1='WETH', tkn0_address='0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84', tkn0_decimals=18, tkn1_address='0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', tkn1_decimals=18, tkn0_key='stETH-fE84', tkn1_key='WETH-6Cc2'), PoolAndTokens(ConfigObj=C, id=46, cid='0xfc4a2b96bc1cbe2515f7c0bb784c80757a75fd8851a2ff4ecbad69cb1869ddfa', last_updated=10, last_updated_block=17632857, descr='bancor_v3 BNT-FF1C/BAT-87EF 0.000', pair_name='BNT-FF1C/BAT-87EF', exchange_name='bancor_v3', fee='0.000', fee_float=0.0, tkn0_balance=517690761518839476171134, tkn1_balance=1017685474801793344376721, z_0=0, y_0=0, A_0=0, B_0=0, z_1=0, y_1=0, A_1=0, B_1=0, sqrt_price_q96=10, tick=10, tick_spacing=0, liquidity=10, address='0x3506424F91fD33084466F402d5D97f05F8e3b4AF', anchor=10, tkn0='BNT', tkn1='BAT', tkn0_address='0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C', tkn0_decimals=18, tkn1_address='0x0D8775F648430679A709E98d2b0Cb6250d2887EF', tkn1_decimals=18, tkn0_key='BNT-FF1C', tkn1_key='BAT-87EF'), PoolAndTokens(ConfigObj=C, id=47, cid='0x1cff7b0388c4532d7def9430fd2685224af19257fbbe12b99d4362f059e7dbd7', last_updated=10, last_updated_block=17632857, descr='bancor_v3 BNT-FF1C/BBS-B430 0.000', pair_name='BNT-FF1C/BBS-B430', exchange_name='bancor_v3', fee='0.000', fee_float=0.0, tkn0_balance=32805970640508809330829, tkn1_balance=1044245786489813870187767, z_0=0, y_0=0, A_0=0, B_0=0, z_1=0, y_1=0, A_1=0, B_1=0, sqrt_price_q96=10, tick=10, tick_spacing=0, liquidity=10, address='0xc12d099be31567add4e4e4d0D45691C3F58f5663', anchor=10, tkn0='BNT', tkn1='BBS', tkn0_address='0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C', tkn0_decimals=18, tkn1_address='0xFe459828c90c0BA4bC8b42F5C5D44F316700B430', tkn1_decimals=18, tkn0_key='BNT-FF1C', tkn1_key='BBS-B430'), PoolAndTokens(ConfigObj=C, id=0, cid='0x7c746d3518854384f7a73a6d2da821454319d305a2af4b7015bbf8734a394e49', last_updated=None, last_updated_block=17632854, descr='uniswap_v3 WETH-6Cc2/USDT-1ec7 500', pair_name='WETH-6Cc2/USDT-1ec7', exchange_name='uniswap_v3', fee=500, fee_float=0.0005, tkn0_balance=None, tkn1_balance=None, z_0=0, y_0=0, A_0=0, B_0=0, z_1=0, y_1=0, A_1=0, B_1=0, sqrt_price_q96=3471749971564948777178138, tick=-200719, tick_spacing=10, liquidity=9272761710152634655, address='0x11b815efB8f581194ae79006d24E0d814B7697F6', anchor=None, tkn0='WETH', tkn1='USDT', tkn0_address='0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', tkn0_decimals=18, tkn1_address='0xdAC17F958D2ee523a2206206994597C13D831ec7', tkn1_decimals=6, tkn0_key='WETH-6Cc2', tkn1_key='USDT-1ec7'), PoolAndTokens(ConfigObj=C, id=1, cid='0x71fa9967e86d546c3103642dd0876f28a32d4665a1a487d18083421cfd3eff02', last_updated=None, last_updated_block=17632853, descr='uniswap_v3 LDO-1B32/WETH-6Cc2 3000', pair_name='LDO-1B32/WETH-6Cc2', exchange_name='uniswap_v3', fee=3000, fee_float=0.003, tkn0_balance=None, tkn1_balance=None, z_0=0, y_0=0, A_0=0, B_0=0, z_1=0, y_1=0, A_1=0, B_1=0, sqrt_price_q96=2621090548713608947690718187, tick=-68179, tick_spacing=60, liquidity=404595565356008216313141, address='0xa3f558aebAecAf0e11cA4b2199cC5Ed341edfd74', anchor=None, tkn0='LDO', tkn1='WETH', tkn0_address='0x5A98FcBEA516Cf06857215779Fd812CA3beF1B32', tkn0_decimals=18, tkn1_address='0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', tkn1_decimals=18, tkn0_key='LDO-1B32', tkn1_key='WETH-6Cc2')]
    state = [pool.__dict__ for pool in pools]
    exchanges = list({ex['exchange_name'] for ex in state})
    mock_qi = QueryInterface(state=state,
                             ConfigObj = C,
                             exchanges=exchanges,
    )
    setup_bot.db = mock_qi
    return setup_bot

# Test bot
@pytest.mark.unittest
def test_bot_and_query_interface(setup_bot):
    assert str(type(setup_bot.db)) == "<class 'fastlane_bot.events.interface.QueryInterface'>"
    pools = setup_bot.db.get_pools()
    assert len(pools) > 0
    assert isinstance(pools, list)
    assert all(
        str(type(pool)).startswith("<class 'fastlane_bot.") for pool in pools
    )
    assert setup_bot.db.get_pools()[0].fee_float is not None, "Incorrect pools.csv file see MockPoolManager data"


@pytest.mark.unittest
def test_setup_curves():
    cc1 = CPC.from_carbon(pair="WETH-6Cc2/USDC-eB48", tkny="WETH-6Cc2", yint=10, y=10, pa=1/2000, pb=1/2010,
                          cid="1701411834604692317316873037158841057285-1", params={"exchange":'carbon_v1'})
    assert iseq(1/2000, cc1.p, cc1.p_max)
    assert iseq(1/2010, cc1.p_min)
    assert cc1.p_convention() == 'WETH per USDC'
    assert cc1.p_min < cc1.p_max



# coding=utf-8
"""
This module contains the tests for the exchanges classes
"""
from dataclasses import asdict
from unittest.mock import MagicMock
from unittest.mock import Mock

import pytest

from fastlane_bot import Bot
from fastlane_bot.events.exchanges import (
    UniswapV2,
    UniswapV3,
    SushiswapV2,
    CarbonV1,
    BancorV3,
)
from fastlane_bot.events.interface import QueryInterface, Token
from fastlane_bot.tools.cpc import ConstantProductCurve as CPC

print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Bot))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV3))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SushiswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonV1))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(BancorV3))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(QueryInterface))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Token))

from fastlane_bot.testing import *

plt.style.use("seaborn-dark")
plt.rcParams["figure.figsize"] = [12, 6]
from fastlane_bot import __VERSION__

require("3.0", __VERSION__)


@pytest.fixture
def qi():
    cfg_mock = Mock()
    cfg_mock.logger = MagicMock()

    qi = QueryInterface(mgr=None, ConfigObj=cfg_mock)

    qi.state = [
        {
            "exchange_name": "uniswap_v2",
            "address": "0x123",
            "tkn0_key": "TKN-0x123",
            "tkn1_key": "TKN-0x456",
            "pair_name": "Pair-0x789",
            "liquidity": 10,
        },
        {
            "exchange_name": "sushiswap_v2",
            "address": "0xabc",
            "tkn0_key": "TKN-0xabc",
            "tkn1_key": "TKN-0xdef",
            "pair_name": "Pair-0xghi",
            "liquidity": 0,
        },
    ]
    return qi


def test_remove_unsupported_exchanges(qi):
    qi.exchanges = ["uniswap_v2"]
    qi.remove_unsupported_exchanges()
    assert len(qi.state) == 1
    assert qi.state[0]["exchange_name"] == "uniswap_v2"


def test_has_balance(qi):
    assert qi.has_balance(qi.state[0], "liquidity") == True
    assert qi.has_balance(qi.state[1], "liquidity") == False


def test_filter_pools(qi):
    assert len(qi.filter_pools("uniswap_v2")) == 1
    assert len(qi.filter_pools("sushiswap_v2")) == 1


def test_cleanup_token_key(qi):
    assert qi.cleanup_token_key("TKN-Hello-0x123") == "TKN_Hello-0x123"
    assert qi.cleanup_token_key("TKN-0x123") == "TKN-0x123"


def test_update_state(qi):
    new_state = [
        {
            "exchange_name": "bancor_v2",
            "address": "0xabc",
            "tkn0_key": "TKN-0xabc",
            "tkn1_key": "TKN-0xdef",
            "pair_name": "Pair-0xghi",
            "liquidity": 10,
        }
    ]
    qi.update_state(new_state)
    assert qi.state == new_state


def test_get_token(qi):
    new_state = [
        {
            "exchange_name": "bancor_v2",
            "address": "0xabc",
            "tkn0_key": "TKN-0x123",
            "tkn1_key": "TKN-0xdef",
            "pair_name": "Pair-0xghi",
            "liquidity": 10,
        }
    ]
    qi.update_state(new_state)
    token = qi.get_token("TKN-0x123")
    assert isinstance(token, Token)
    assert token.key == "TKN-0x123"


def test_get_pool(qi):
    new_state = [
        {
            "last_updated_block": 17614344,
            "address": "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1",
            "exchange_name": "carbon_v1",
            "tkn0_address": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "tkn1_address": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "tkn0_symbol": "ETH",
            "tkn1_symbol": "USDC",
            "tkn0_decimals": 18,
            "tkn1_decimals": 6,
            "cid": 1701411834604692317316873037158841057365,
            "tkn0_key": "ETH-EEeE",
            "tkn1_key": "USDC-eB48",
            "pair_name": "ETH-EEeE/USDC-eB48",
            "fee_float": 0.002,
            "fee": "0.002",
            "descr": "carbon_v1 ETH-EEeE/USDC-eB48 0.002",
            "y_0": 9882507039899549,
            "y_1": 0,
            "z_0": 9882507039899549,
            "z_1": 17936137,
            "A_0": 0,
            "A_1": 99105201,
            "B_0": 0,
            "B_1": 11941971885,
        }
    ]
    qi.update_state(new_state)
    pool = qi.get_pool(cid=1701411834604692317316873037158841057365)
    assert asdict(pool)["cid"] == 1701411834604692317316873037158841057365



# coding=utf-8
"""
This module contains the tests for the exchanges classes
"""

from fastlane_bot import Bot
from fastlane_bot.events.pools import (
    SushiswapV2Pool,
    UniswapV2Pool,
    UniswapV3Pool,
    BancorV3Pool,
    CarbonV1Pool,
)
from fastlane_bot.tools.cpc import ConstantProductCurve as CPC
from typing import List
from web3.datastructures import AttributeDict
from web3.types import HexBytes
import pytest

from fastlane_bot.events.utils import filter_latest_events, complex_handler

print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Bot))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV2Pool))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV3Pool))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SushiswapV2Pool))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonV1Pool))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(BancorV3Pool))
from fastlane_bot.testing import *

plt.style.use("seaborn-dark")
plt.rcParams["figure.figsize"] = [12, 6]
from fastlane_bot import __VERSION__

require("3.0", __VERSION__)


@pytest.fixture
def mocked_mgr():
    class MockManager:
        def pool_type_from_exchange_name(self, exchange_name):
            class MockPoolType:
                def unique_key(self):
                    return "address"

            return MockPoolType()

        def exchange_name_from_event(self, event):
            return "uniswap_v2"

    mocked_mgr = MockManager()
    return mocked_mgr


@pytest.fixture
def mock_events() -> List[List[AttributeDict]]:
    event1 = AttributeDict(
        {
            "args": {"reserve0": 100, "reserve1": 100},
            "event": "Sync",
            "address": "0xabc",
            "blockNumber": 5,
            "transactionIndex": 0,
            "logIndex": 0,
        }
    )
    event2 = AttributeDict(
        {
            "args": {"reserve0": 200, "reserve1": 200},
            "event": "Sync",
            "address": "0xabc",
            "blockNumber": 10,
            "transactionIndex": 1,
            "logIndex": 1,
        }
    )
    event3 = AttributeDict(
        {
            "args": {"reserve0": 300, "reserve1": 300},
            "event": "Sync",
            "address": "0xdef",
            "blockNumber": 7,
            "transactionIndex": 1,
            "logIndex": 1,
        }
    )
    mock_events = [[event1, event2, event3]]
    return mock_events


def test_filter_latest_events(mocked_mgr, mock_events):
    result = filter_latest_events(mocked_mgr, mock_events)

    # Check that we only have one event per pool
    assert len(result) == len({event["address"] for event in result})

    # Select first pool's address
    pool_address = result[0]["address"]

    # Get all events for this pool
    pool_events = [
        event for event in mock_events[0] if event["address"] == pool_address
    ]

    # Check that the event is indeed the latest one for this pool
    assert result[0] == max(pool_events, key=lambda e: e["blockNumber"])


def test_complex_handler():
    # AttributeDict case
    attribute_dict = AttributeDict({"a": 1, "b": 2})
    assert complex_handler(attribute_dict) == {"a": 1, "b": 2}

    # HexBytes case
    hex_bytes = HexBytes(b"hello")
    assert complex_handler(hex_bytes) == "0x68656c6c6f"

    # dict case
    dictionary = {"a": 1, "b": HexBytes(b"hello"), "c": AttributeDict({"d": 4})}
    assert complex_handler(dictionary) == {"a": 1, "b": "0x68656c6c6f", "c": {"d": 4}}

    # list case
    list_ = [1, HexBytes(b"hello"), AttributeDict({"d": 4})]
    assert complex_handler(list_) == [1, "0x68656c6c6f", {"d": 4}]

    # set case
    set_ = {1, 2, 3}
    assert complex_handler(set_) == [1, 2, 3]

    # other case
    assert complex_handler(123) == 123
    assert complex_handler("abc") == "abc"
